@@ -24,6 +24,17 @@
 #include "buffy.h"
 
 
+
+
+
+
+
+
+
+
+
+
+
 #include <ctype.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -59,29 +70,62 @@ extern const char *ReleaseDate;
 
 void index_make_entry (char *s, size_t l, MUTTMENU *menu, int num)
 {
-  int flag = M_MAKEPRINT;
-  HEADER *h = Context->hdrs[Context->v2r[num]];
+  format_flag flag = M_FORMAT_MAKEPRINT;
+  int reverse = Sort & SORT_REVERSE, edgemsgno;
+  HEADER *tmp, *h = Context->hdrs[Context->v2r[num]];
 
-  if ((Sort & SORT_MASK) == SORT_THREADS &&
-      Context->msgcount == Context->vcount &&
-      h->tree)
+  if ((Sort & SORT_MASK) == SORT_THREADS && h->tree)
   {
-    if (h->display_subject)
-      flag = M_FORCESUBJ | M_MAKEPRINT;
-    else if (Sort & SORT_REVERSE)
-    {
-      if ((h->parent->msgno > menu->top + menu->pagelen - 1) && 
-	  (!h->prev || (h->prev->msgno > menu->top + menu->pagelen - 1)))
-	flag = M_FORCESUBJ | M_MAKEPRINT;
-    }
-    else if ((h->parent->msgno < menu->top) &&
-	     (!h->prev || (h->prev->msgno < menu->top)))
-      flag = M_FORCESUBJ | M_MAKEPRINT;
+    flag |= M_FORMAT_TREE; /* display the thread tree */
 
-    flag |= M_TREE; /* display the thread tree */
+    if (h->display_subject)
+      flag |= M_FORMAT_FORCESUBJ;
+    else
+    {
+      if (reverse)
+      {
+       if (menu->top + menu->pagelen > menu->max)
+         edgemsgno = Context->v2r[menu->max - 1];
+       else
+         edgemsgno = Context->v2r[menu->top + menu->pagelen - 1];
+      }
+      else
+       edgemsgno = Context->v2r[menu->top];
+
+      for (tmp = h->parent; tmp; tmp = tmp->parent)
+      {
+	if ((reverse && tmp->msgno > edgemsgno)
+	    || (!reverse && tmp->msgno < edgemsgno))
+	{
+	  flag |= M_FORMAT_FORCESUBJ;
+	  break;
+	}
+	else if (tmp->virtual >= 0)
+	  break;
+      }
+      if ((flag & M_FORMAT_FORCESUBJ) && h->prev)
+      {
+	for (tmp = h->prev; tmp; tmp = tmp->prev)
+	{
+	  if ((reverse && tmp->msgno > edgemsgno)
+	      || (!reverse && tmp->msgno < edgemsgno))
+	    break;
+	  else if (tmp->virtual >= 0)
+	  {
+	    flag &= ~M_FORMAT_FORCESUBJ;
+	    break;
+	  }
+	}
+      }
+    }
   }
 
-  _mutt_make_string (s, l, HdrFmt, h, flag);
+  _mutt_make_string (s, l, NONULL (HdrFmt), Context, h, flag);
+}
+
+int index_color (int index_no)
+{
+  return Context->hdrs[Context->v2r[index_no]]->pair;
 }
 
 static int ci_next_undeleted (int msgno)
@@ -202,6 +246,7 @@ void mutt_index_menu (void)
   menu->offset = 1;
   menu->pagelen = LINES - 3;
   menu->make_entry = index_make_entry;
+  menu->color = index_color;
   menu->current = ci_first_message ();
   menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_MAIN, IndexHelp);
   
@@ -244,6 +289,29 @@ void mutt_index_menu (void)
 	    oldcount = 0; /* invalid message number! */
 	}
 
+	/* We are in a limited view. Check if the new message(s) satisfy
+	 * the limit criteria. If they do, set their virtual msgno so that
+	 * they will be visible in the limited view */
+        if (Context->pattern)
+	{
+          #define this_body Context->hdrs[i]->content
+	  if (oldcount)
+          for (i = oldcount; i < Context->msgcount; i++)
+	  {
+            if (mutt_pattern_exec (Context->limit_pattern,
+	                           M_MATCH_FULL_ADDRESS, 
+	                           Context, Context->hdrs[i]))
+            {
+	     Context->hdrs[i]->virtual = Context->vcount;
+	     Context->v2r[Context->vcount] = i;
+	     Context->vcount++;
+	     Context->vsize+=this_body->length + this_body->offset -
+                             this_body->hdr_offset;
+            }
+	  }
+          #undef this_body
+        }
+
 	/* if the mailbox was reopened, need to rethread from scratch */
 	mutt_sort_headers (Context, (i == M_REOPENED));
 
@@ -322,7 +390,7 @@ void mutt_index_menu (void)
 
       if (menu->redraw & REDRAW_STATUS) 
       {
-	menu_status_line (buf, sizeof (buf), menu, Status);
+	menu_status_line (buf, sizeof (buf), menu, NONULL (Status));
 	CLEARLINE (option (OPTSTATUSONTOP) ? 0 : LINES-2);
 	SETCOLOR (MT_COLOR_STATUS);
 	printw ("%-*.*s", COLS, COLS, buf);
@@ -465,7 +533,7 @@ void mutt_index_menu (void)
 	    !buf[0])
 	  break;
 
-	if (! isdigit (buf[0]))
+	if (! isdigit ((unsigned char) buf[0]))
 	{
 	  mutt_error ("Argument must be a message number.");
 	  break;
@@ -533,6 +601,18 @@ void mutt_index_menu (void)
 	menu->redraw = REDRAW_FULL;
 	break;
 
+      case OP_MAIN_SHOW_LIMIT:
+        CHECK_MSGCOUNT;
+	if (!Context->pattern)
+	   mutt_message ("No limit pattern is in effect.");
+	else
+	{
+	   char buf[STRING];
+	   snprintf (buf, sizeof(buf), "Limit: %s",Context->pattern);
+           mutt_message ("%s", buf);
+	}
+        break;
+
       case OP_MAIN_LIMIT:
 
 	CHECK_MSGCOUNT;
@@ -554,15 +634,19 @@ void mutt_index_menu (void)
 	  else
 	    menu->current = 0;
 	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	  mutt_linearize_tree (Context, 0);
 	}
 	break;	  
 
       case OP_QUIT:
 
-	if (!Context || mx_close_mailbox (Context) == 0)
-	  done = 1;
-	else
-	  menu->redraw = REDRAW_FULL; /* new mail arrived? */
+	if (query_quadoption (OPT_QUIT, "Quit Mutt?") == M_YES)
+	{ 
+	  if (!Context || mx_close_mailbox (Context) == 0)
+	    done = 1;
+	  else
+	    menu->redraw = REDRAW_FULL; /* new mail arrived? */
+	}
 	break;
 
       case OP_REDRAW:
@@ -721,7 +805,10 @@ void mutt_index_menu (void)
 	}
 
         if (Context)
-	  strfcpy (LastFolder, Context->path, sizeof (LastFolder));
+        {
+	  FREE (&LastFolder);
+	  LastFolder = safe_strdup (Context->path);
+	}
 
 	if (Context)
 	{
@@ -792,7 +879,9 @@ void mutt_index_menu (void)
 
       case OP_EXIT:
 
-	if (menu->menu == MENU_MAIN)
+	if ((menu->menu == MENU_MAIN)
+	    && (query_quadoption (OPT_QUIT, 
+				  "Exit Mutt without saving?") == M_YES))
 	{
 	  if (Context)
 	  {
@@ -1126,7 +1215,12 @@ void mutt_index_menu (void)
 
       case OP_CREATE_ALIAS:
 
-	mutt_create_alias (Context && Context->vcount ? CURHDR->env : NULL);
+	mutt_create_alias (Context && Context->vcount ? CURHDR->env : NULL, NULL);
+	MAYBE_REDRAW (menu->redraw);
+	break;
+
+      case OP_QUERY:
+	mutt_query_menu (NULL, 0);
 	MAYBE_REDRAW (menu->redraw);
 	break;
 
@@ -1190,7 +1284,7 @@ void mutt_index_menu (void)
       case OP_ENTER_COMMAND:
 
 	mutt_enter_command ();
-
+	mutt_check_rescore (Context);
 	if (option (OPTNEEDRESORT) && Context && Context->msgcount)
 	{
 	  menu->oldcurrent = CURHDR->index;
@@ -1204,12 +1298,10 @@ void mutt_index_menu (void)
 	  }
 	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
 	}
-
 	if (option (OPTFORCEREDRAWINDEX))
 	  menu->redraw = REDRAW_FULL;
 	unset_option (OPTFORCEREDRAWINDEX);
 	unset_option (OPTFORCEREDRAWPAGER);
-	
 	break;
 
       case OP_FORWARD_MESSAGE:
@@ -1220,6 +1312,10 @@ void mutt_index_menu (void)
 	break;
 
 
+
+
+
+
       case OP_GROUP_REPLY:
 
 	CHECK_MSGCOUNT;
@@ -1241,6 +1337,18 @@ void mutt_index_menu (void)
 	break;
 
 
+
+
+
+
+
+
+
+
+
+
+
+
       case OP_PIPE:
 
 	CHECK_MSGCOUNT;
@@ -1367,6 +1475,8 @@ void mutt_index_menu (void)
       case OP_VIEW_ATTACHMENTS:
 	CHECK_MSGCOUNT;
 	mutt_view_attachments (CURHDR);
+	if (CURHDR->attach_del)
+	  Context->changed = 1;
 	menu->redraw = REDRAW_FULL;
 	break;
 
@@ -1387,3 +1497,28 @@ void mutt_index_menu (void)
 
   mutt_menuDestroy (&menu);
 }
+
+void mutt_set_header_color (CONTEXT *ctx, HEADER *curhdr)
+{
+  COLOR_LINE *color;
+
+  if (!curhdr)
+    return;
+  
+  for (color = ColorIndexList; color; color = color->next)
+   if (mutt_pattern_exec (color->color_pattern, M_MATCH_FULL_ADDRESS, ctx, curhdr))
+   {
+      curhdr->pair = color->pair;
+      return;
+   }
+  curhdr->pair = ColorDefs[MT_COLOR_NORMAL];
+}
+
+void mutt_cache_index_colors (CONTEXT *ctx)
+{
+ int i;
+
+ if (ctx)
+   for (i = 0; i < ctx->msgcount; i++)
+     mutt_set_header_color (ctx, ctx->hdrs[i]);
+}
