@@ -62,74 +62,127 @@ int Index_64[128] = {
 
 void mutt_decode_xbit (STATE *s, long len, int istext)
 {
-  int linelen;
-  char buffer[LONG_STRING];
-
+  int c;
+  int lbreak = 1;
+  
   if (istext)
   {
-    while (len > 0)
+    while ((c = fgetc(s->fpin)) != EOF && len--)
     {
-      if (fgets (buffer, LONG_STRING, s->fpin) == 0) return;
-      linelen = strlen (buffer);
-      len -= linelen;
-      if (linelen >= 2 && buffer[linelen-2] == '\r')
+      if(lbreak && s->prefix)
       {
-	buffer[linelen-2] = '\n';
-	buffer[linelen-1] = 0;
+	state_puts(s->prefix, s);
+	lbreak = 0;
       }
-      if (s->prefix) state_puts (s->prefix, s);
-      state_puts (buffer, s);
+	  
+      if (c == '\r' && len)
+      {
+	int ch;
+	
+	if((ch = fgetc(s->fpin)) != '\n')
+	  ungetc(ch, s->fpin);
+	else
+	{
+	  c = ch;
+	  len--;
+	}
+	
+      }
+      fputc(c, s->fpout);
+      if(c == '\n')
+	lbreak = 1;
     }
   }
   else
     mutt_copy_bytes (s->fpin, s->fpout, len);
 }
 
+static int handler_state_fgetc(STATE *s)
+{
+  int ch;
+  
+  if((ch = fgetc(s->fpin)) == EOF)
+  {
+    dprint(1, (debugfile, "handler_state_fgetc: unexpected EOF.\n"));
+    state_puts ("[-- Error: unexpected end of file! --]\n", s);
+  }
+  return ch;
+}
+
 void mutt_decode_quoted (STATE *s, long len, int istext)
 {
-  char *c, buffer[LONG_STRING];
-  int ch, soft = 0;
+  int ch, lbreak = 1;
 
   while (len > 0)
   {
-    if (fgets (buffer, LONG_STRING, s->fpin) == NULL)
-    {
-      dprint (1, (debugfile, "mutt_decode_quoted: unexpected EOF.\n"));
-      state_puts ("[-- Error: unexpected end of file! --]\n", s);
+    if ((ch = handler_state_fgetc(s)) == EOF)
       break;
-    }
-    c = buffer;
-    len -= strlen (buffer);
-    if (s->prefix && !soft) state_puts (s->prefix, s);
-    soft = 0;
-    while (*c)
+
+    len--;
+    
+    if (s->prefix && lbreak)
+      state_puts (s->prefix, s);
+    
+    lbreak = 0;
+    if (ch == '=')
     {
-      if (*c == '=')
+      int ch1, ch2;
+      
+      if(!len || (ch1 = handler_state_fgetc(s)) == EOF)
+	break;
+
+      len--;
+      
+      if (ch1 == '\n' || ch1 == '\r' || ch1 == ' ' || ch1 == '\t')
       {
-        if (c[1] == '\n' || c[1] == '\r' || c[1] == ' ' || c[1] == '\t')
+	/* Skip whitespace at the end of the line since MIME does not
+	 * allow for it
+	 */
+	if(ch1 != '\n')
 	{
-          /* Skip whitespace at the end of the line since MIME does not
-           * allow for it
-	   */
-          soft = 1;
-          break;
-        }
-        ch = hexval ((int) c[1]) << 4;
-        ch |= hexval ((int) c[2]);
-        state_putc (ch, s);
-        c += 3;
-      }
-      else if (istext && c[0] == '\r' && c[1] == '\n')
-      {
-        state_putc ('\n', s);
-        break;
+	  while(len && (ch1 = handler_state_fgetc(s)) != EOF)
+	  {
+	    len--;
+	    if(ch1 == '\n')
+	      break;
+	  }
+	}
+
+	if(ch1 == EOF)
+	  break;
+
+	ch = EOF;
+
       }
       else
       {
-        state_putc (*c, s);
-        c++;
+	if(!len || (ch2 = handler_state_fgetc(s)) == EOF)
+	  break;
+
+	len--;
+	
+        ch = hexval (ch1) << 4;
+        ch |= hexval (ch2);
       }
+    } /* ch == '=' */
+    else if (istext && ch == '\r')
+    {
+      int ch1;
+
+      if((ch1 =fgetc(s->fpin)) == '\n')
+      {
+	ch = ch1;
+	len--;
+      }
+      else
+	ungetc(ch1, s->fpin);
     }
+
+    if(ch != EOF)
+      state_putc (ch, s);
+
+    if(ch == '\n')
+      lbreak = 1;
   }
 }
 
@@ -652,6 +705,7 @@ void text_enriched_handler (BODY *a, STATE *s)
 
       case ST_EOF :
 	enriched_putc ('\0', &stte);
+        enriched_flush (&stte, 1);
 	state = DONE;
 	break;
 
@@ -676,10 +730,45 @@ void alternative_handler (BODY *a, STATE *s)
 {
   BODY *choice = NULL;
   BODY *b;
+  LIST *t;
   char buf[STRING];
   int type = 0;
 
-  /* First, look for an autoviewable type */
+  /* First, search list of prefered types */
+  t = AlternativeOrderList;
+  while (t && !choice)
+  {
+    if (a && a->parts) 
+      b = a->parts;
+    else
+      b = a;
+    while (b && !choice)
+    {
+      int i;
+
+      /* catch base only matches */
+      i = strlen (t->data) - 1;
+      if (!strchr(t->data, '/') || 
+	  (i > 0 && t->data[i-1] == '/' && t->data[i] == '*'))
+      {
+	if (!strcasecmp(t->data, TYPE(b->type)))
+	{
+	  choice = b;
+	}
+      }
+      else
+      {
+	snprintf (buf, sizeof (buf), "%s/%s", TYPE (b->type), b->subtype);
+	if (!strcasecmp(t->data, buf))
+	{
+	  choice = b;
+	}
+      }
+      b = b->next;
+    }
+    t = t->next;
+  }
+  /* Next, look for an autoviewable type */
   if (a && a->parts) 
     b = a->parts;
   else
@@ -815,6 +904,12 @@ int mutt_can_decode (BODY *a)
     return (1);
   else if (a->type == TYPEMULTIPART)
   {
+
+
+
+
+
+
     {
       BODY *p;
 
@@ -825,6 +920,12 @@ int mutt_can_decode (BODY *a)
       }
     }
   }
+
+
+
+
+
+
   return (0);
 }
 
@@ -1071,14 +1172,28 @@ void mutt_body_handler (BODY *b, STATE *s)
   else if (b->type == TYPEMULTIPART)
   {
 
+
+
+
+
+
     if (strcasecmp ("alternative", b->subtype) == 0)
       handler = alternative_handler;
 
 
+
+
+
+
     if (!handler)
       handler = multipart_handler;
   }
 
+
+
+
+
+
   if (plaintext || handler)
   {
     fseek (s->fpin, b->offset, 0);
