@@ -39,7 +39,7 @@ BODY *mutt_new_body (void)
   return (p);
 }
 
-BODY* mutt_dup_body (BODY *b)
+BODY *mutt_dup_body (BODY *b)
 {
   BODY *bn;
 
@@ -214,7 +214,27 @@ char *mutt_expand_path (char *s, size_t slen)
     }
   }
   else if (*s == '=' || *s == '+')
-    snprintf (p, sizeof (p), "%s/%s", Maildir, s + 1);
+    snprintf (p, sizeof (p), "%s/%s", NONULL (Maildir), s + 1);
+  else if (*s == '@')
+  {
+    /* elm compatibility, @ expands alias to user name */
+    HEADER *h;
+    ADDRESS *alias;
+
+    alias = mutt_lookup_alias (s + 1);
+    if (alias != NULL)
+    {
+      h = mutt_new_header();
+      h->env = mutt_new_envelope();
+      h->env->from = h->env->to = alias;
+      mutt_default_save (p, sizeof (p), h);
+      h->env->from = h->env->to = NULL;
+      mutt_free_header (&h);
+      /* Avoid infinite recursion if the resulting folder starts with '@' */
+      if (*p != '@')
+	mutt_expand_path (p, sizeof (p));
+    }
+  }
   else
   {
     if (*s == '>')
@@ -228,7 +248,7 @@ char *mutt_expand_path (char *s, size_t slen)
     else
       return s;
 
-    if (!*q)
+    if (!q)
       return s;
     snprintf (p, sizeof (p), "%s%s", q, s + 1);
   }
@@ -370,6 +390,10 @@ int mutt_needs_mailcap (BODY *m)
       break;
 
 
+
+
+
+
     case TYPEMULTIPART:
     case TYPEMESSAGE:
 
@@ -391,6 +415,10 @@ int mutt_is_text_type (int t, char *s)
   }
 
 
+
+
+
+
   return 0;
 }
 
@@ -424,7 +452,7 @@ void mutt_tabs_to_spaces (char *s)
 
 void mutt_mktemp (char *s)
 {
-  snprintf (s, _POSIX_PATH_MAX, "%s/mutt-%s-%d-%d", Tempdir, Hostname, (int) getpid (), Counter++);
+  snprintf (s, _POSIX_PATH_MAX, "%s/mutt-%s-%d-%d", NONULL (Tempdir), Hostname, (int) getpid (), Counter++);
   unlink (s);
 }
 
@@ -494,7 +522,7 @@ void mutt_pretty_mailbox (char *s)
   }
   *q = 0;
 
-  if (strncmp (s, Maildir, (len = strlen (Maildir))) == 0 && s[len] == '/')
+  if (strncmp (s, NONULL (Maildir), (len = strlen (NONULL (Maildir)))) == 0 && s[len] == '/')
   {
     *s++ = '=';
     strcpy (s, s + len);
@@ -700,22 +728,29 @@ void mutt_pretty_size (char *s, size_t len, long n)
   }
 }
 
+void mutt_save_path (char *d, size_t dsize, ADDRESS *a)
+{
+  if (a && a->mailbox)
+  {
+    strfcpy (d, a->mailbox, dsize);
+    if (!option (OPTSAVEADDRESS))
+    {
+      char *p;
+
+      if ((p = strpbrk (d, "%@")))
+	*p = 0;
+    }
+    mutt_strlower (d);
+  }
+  else
+    *d = 0;
+}
+
 void mutt_safe_path (char *s, size_t l, ADDRESS *a)
 {
   char *p;
 
-  if (a && a->mailbox)
-  {
-    strfcpy (s, a->mailbox, l);
-    if (!option (OPTSAVEADDRESS))
-    {
-    if ((p = strpbrk (s, "%@")))
-      *p = 0;
-    }
-  }
-  else
-    *s = 0;
-  mutt_strlower (s);
+  mutt_save_path (s, l, a);
   for (p = s; *p; p++)
     if (*p == '/' || ISSPACE (*p) || !IsPrint ((unsigned char) *p))
       *p = '_';
@@ -776,3 +811,231 @@ mutt_substrcpy (char *dest, const char *beg, const char *end, size_t destlen)
   dest[len] = 0;
   return dest;
 }
+
+char *mutt_substrdup (const char *begin, const char *end)
+{
+  size_t len;
+  char *p;
+
+  len = end - begin;
+  p = safe_malloc (len + 1);
+  memcpy (p, begin, len);
+  p[len] = 0;
+  return p;
+}
+
+void mutt_FormatString (char *dest,		/* output buffer */
+			size_t destlen,		/* output buffer len */
+			const char *src,	/* template string */
+			format_t *callback,	/* callback for processing */
+			unsigned long data,	/* callback data */
+			format_flag flags)	/* callback flags */
+{
+  char prefix[SHORT_STRING], buf[LONG_STRING], *cp, *wptr = dest, ch;
+  char ifstring[SHORT_STRING], elsestring[SHORT_STRING];
+  size_t wlen = 0, count, len;
+
+  destlen--; /* save room for the terminal \0 */
+  while (*src && wlen < destlen)
+  {
+    if (*src == '%')
+    {
+      if (*++src == '%')
+      {
+	*wptr++ = '%';
+	wlen++;
+	src++;
+	continue;
+      }
+
+      if (*src == '?')
+      {
+	flags |= M_FORMAT_OPTIONAL;
+	src++;
+      }
+      else
+      {
+	flags &= ~M_FORMAT_OPTIONAL;
+
+	/* eat the format string */
+	cp = prefix;
+	count = 0;
+	while (count < sizeof (prefix) &&
+	       (isdigit ((unsigned char) *src) || *src == '.' || *src == '-'))
+	{
+	  *cp++ = *src++;
+	  count++;
+	}
+	*cp = 0;
+      }
+
+      if (!*src)
+	break; /* bad format */
+
+      ch = *src++; /* save the character to switch on */
+
+      if (flags & M_FORMAT_OPTIONAL)
+      {
+        if (*src != '?')
+          break; /* bad format */
+        src++;
+
+        /* eat the `if' part of the string */
+        cp = ifstring;
+	count = 0;
+        while (count < sizeof (ifstring) && *src && *src != '?' && *src != '&')
+	{
+          *cp++ = *src++;
+	  count++;
+	}
+        *cp = 0;
+
+	/* eat the `else' part of the string (optional) */
+	if (*src == '&')
+	  src++; /* skip the & */
+	cp = elsestring;
+	count = 0;
+	while (count < sizeof (elsestring) && *src && *src != '?')
+	{
+	  *cp++ = *src++;
+	  count++;
+	}
+	*cp = 0;
+
+	if (!*src)
+	  break; /* bad format */
+
+        src++; /* move past the trailing `?' */
+      }
+
+      /* handle generic cases first */
+      if (ch == '>')
+      {
+	/* right justify to EOL */
+	ch = *src++; /* pad char */
+	/* calculate space left on line.  if we've already written more data
+	   than will fit on the line, ignore the rest of the line */
+	count = (COLS < destlen ? COLS : destlen);
+	if (count > wlen)
+	{
+	  count -= wlen; /* how many chars left on this line */
+	  mutt_FormatString (buf, sizeof (buf), src, callback, data, flags);
+	  len = strlen (buf);
+	  if (count > len)
+	  {
+	    count -= len; /* how many chars to pad */
+	    memset (wptr, ch, count);
+	    wptr += count;
+	    wlen += count;
+	  }
+	  if (len + wlen > destlen)
+	    len = destlen - wlen;
+	  memcpy (wptr, buf, len);
+	  wptr += len;
+	  wlen += len;
+	}
+	break; /* skip rest of input */
+      }
+      else if (ch == '|')
+      {
+	/* pad to EOL */
+	ch = *src++;
+	if (destlen > COLS)
+	  destlen = COLS;
+	if (destlen > wlen)
+	{
+	  count = destlen - wlen;
+	  memset (wptr, ch, count);
+	  wptr += count;
+	}
+	break; /* skip rest of input */
+      }
+      else
+      {
+	/* use callback function to handle this case */
+	src = callback (buf, sizeof (buf), ch, src, prefix, ifstring, elsestring, data, flags);
+
+	if ((len = strlen (buf)) + wlen > destlen)
+	{
+	  if ((len = destlen - wlen) < 0)
+	    len = 0;
+	}
+	memcpy (wptr, buf, len);
+	wptr += len;
+	wlen += len;
+      }
+    }
+    else if (*src == '\\')
+    {
+      if (!*++src)
+	break;
+      switch (*src)
+      {
+	case 'n':
+	  *wptr = '\n';
+	  break;
+	case 't':
+	  *wptr = '\t';
+	  break;
+	case 'r':
+	  *wptr = '\r';
+	  break;
+	case 'f':
+	  *wptr = '\f';
+	  break;
+	case 'v':
+	  *wptr = '\v';
+	  break;
+	default:
+	  *wptr = *src;
+	  break;
+      }
+      src++;
+      wptr++;
+      wlen++;
+    }
+    else
+    {
+      *wptr++ = *src++;
+      wlen++;
+    }
+  }
+  *wptr = 0;
+
+  if (flags & M_FORMAT_MAKEPRINT)
+  {
+    /* Make sure that the string is printable by changing all non-printable
+       chars to dots, or spaces for non-printable whitespace */
+    for (cp = dest ; *cp ; cp++)
+      if (!IsPrint (*cp) &&
+	  !((flags & M_FORMAT_TREE) && (*cp <= M_TREE_MAX)))
+	*cp = isspace ((unsigned char) *cp) ? ' ' : '.';
+  }
+}
+
+/* This function allows the user to specify a command to read stdout from in
+   place of a normal file.  If the last character in the string is a pipe (|),
+   then we assume it is a commmand to run instead of a normal file. */
+FILE *mutt_open_read (const char *path, pid_t *thepid)
+{
+  FILE *f;
+  int len = strlen (path);
+
+  if (path[len - 1] == '|')
+  {
+    /* read from a pipe */
+
+    char *s = safe_strdup (path);
+
+    s[len - 1] = 0;
+    endwin ();
+    *thepid = mutt_create_filter (s, NULL, &f, NULL);
+    free (s);
+  }
+  else
+  {
+    f = fopen (path, "r");
+    *thepid = -1;
+  }
+  return (f);
+}
