@@ -27,6 +27,15 @@
 #include "rfc822.h"
 #include "hash.h"
 
+/* nifty trick I stole from ELM 2.5alpha. */
+#ifdef MAIN_C
+#define WHERE 
+#define INITVAL(x) = x
+#else
+#define WHERE extern
+#define INITVAL(x) 
+#endif
+
 #define TRUE 1
 #define FALSE 0
 
@@ -50,29 +59,42 @@
 #define CH_TXTPLAIN	(1<<11) /* generate text/plain MIME headers */
 
 /* flags for mutt_enter_string() */
-#define  M_ALIAS 1      /* do alias "completion" by calling up the alias-menu */
-#define  M_FILE  (1<<1) /* do file completion */
-#define  M_EFILE (1<<2) /* do file completion, plus incoming folders */
-#define  M_CMD   (1<<3) /* do completion on previous word */
-#define  M_PASS  (1<<4) /* password mode (no echo) */
-#define  M_CLEAR (1<<5) /* clear input if printable character is pressed */
+#define  M_ALIAS   1      /* do alias "completion" by calling up the alias-menu */
+#define  M_FILE    (1<<1) /* do file completion */
+#define  M_EFILE   (1<<2) /* do file completion, plus incoming folders */
+#define  M_CMD     (1<<3) /* do completion on previous word */
+#define  M_PASS    (1<<4) /* password mode (no echo) */
+#define  M_CLEAR   (1<<5) /* clear input if printable character is pressed */
+#define  M_COMMAND (1<<6) /* do command completion */
 
-/* flags for mutt_extract_token() */
-#define M_EQUAL		1	/* treat '=' as a special */
-#define M_CONDENSE	(1<<1)	/* ^(char) to control chars (macros) */
-#define M_SPACE		(1<<2)  /* don't treat whitespace as a terminator */
-#define M_QUOTE		(1<<3)	/* don't interpret quotes */
-#define M_PATTERN	(1<<4)	/* !)|~ are terminators (for patterns) */
-#define M_COMMENT	(1<<5)	/* don't reap comments */
-#define M_NULL		(1<<6)  /* don't return NULL when \0 is reached */
+/* flags for mutt_get_token() */
+#define M_TOKEN_EQUAL		1	/* treat '=' as a special */
+#define M_TOKEN_CONDENSE	(1<<1)	/* ^(char) to control chars (macros) */
+#define M_TOKEN_SPACE		(1<<2)  /* don't treat whitespace as a term */
+#define M_TOKEN_QUOTE		(1<<3)	/* don't interpret quotes */
+#define M_TOKEN_PATTERN		(1<<4)	/* !)|~ are terms (for patterns) */
+#define M_TOKEN_COMMENT		(1<<5)	/* don't reap comments */
+#define M_TOKEN_SEMICOLON	(1<<6)	/* don't treat ; as special */
+
+typedef struct
+{
+  char *data;	/* pointer to data */
+  char *dptr;	/* current read/write position */
+  size_t dsize;	/* length of data */
+  int destroy;	/* destroy `data' when done? */
+} BUFFER;
 
 /* flags for _mutt_system() */
 #define M_DETACH_PROCESS	1	/* detach subprocess from group */
 
-/* flags for _mutt_make_string() */
-#define M_FORCESUBJ	(1<<0) /* print the subject even if it didn't change */
-#define M_TREE		(1<<1) /* draw the thread tree */
-#define M_MAKEPRINT	(1<<2) /* make sure that all chars are printable */
+/* flags for mutt_FormatString() */
+typedef enum
+{
+  M_FORMAT_FORCESUBJ	= (1<<0), /* print the subject even if unchanged */
+  M_FORMAT_TREE		= (1<<1), /* draw the thread tree */
+  M_FORMAT_MAKEPRINT	= (1<<2), /* make sure that all chars are printable */
+  M_FORMAT_OPTIONAL	= (1<<3)
+} format_flag;
 
 /* types for mutt_add_hook() */
 #define M_FOLDERHOOK	1
@@ -81,6 +103,18 @@
 #define M_FCCHOOK	(1<<3)
 #define M_SAVEHOOK	(1<<4)
 
+/* tree characters for linearize_tree and print_enriched_string */
+#define M_TREE_LLCORNER		1
+#define M_TREE_ULCORNER		2
+#define M_TREE_LTEE		3
+#define M_TREE_HLINE		4
+#define M_TREE_VLINE		5
+#define M_TREE_SPACE		6
+#define M_TREE_RARROW		7
+#define M_TREE_STAR		8
+#define M_TREE_HIDDEN		9
+#define M_TREE_MAX		10
+
 enum
 {
   /* modes for mutt_view_attachment() */
@@ -118,9 +152,11 @@ enum
   M_ID,
   M_BODY,
   M_HEADER,
+  M_WHOLE_MSG,
   M_SENDER,
   M_MESSAGE,
   M_SCORE,
+  M_SIZE,
   M_REFERENCE,
   M_RECIPIENT,
   M_LIST,
@@ -156,9 +192,11 @@ enum
   OPT_PRINT,
   OPT_INCLUDE,
   OPT_DELETE,
+  OPT_MIMEFWD,
   OPT_MOVE,
   OPT_COPY,
   OPT_POSTPONE,
+  OPT_QUIT,
   OPT_REPLYTO,
   OPT_ABORT,
   OPT_RECALL,
@@ -202,10 +240,8 @@ enum
   OPTSAVENAME,
   OPTTHOROUGHSRC,
   OPTTILDE,
-  OPTMIMEFWD,
   OPTMARKERS,
   OPTFCCATTACH,
-  OPTATTACHSPLIT,
   OPTPIPESPLIT,
   OPTPIPEDECODE,
   OPTREADONLY,
@@ -220,6 +256,7 @@ enum
   OPTCHECKNEW,
   OPTFASTREPLY,
   OPTWAITKEY,
+  OPTWRAPSEARCH,
   OPTIGNORELISTREPLYTO,
   OPTSAVEADDRESS,
   OPTSUSPEND,
@@ -228,8 +265,10 @@ enum
   OPTUSEFROM,
   OPTUSE8BITMIME,
   OPTFORWDECODE,
+  OPTMIMEFORWDECODE,
   OPTFORWQUOTE,
   OPTBEEPNEW,
+  OPTFOLLOWUPTO,
   OPTMENUSCROLL,	/* scroll menu instead of implicit next-page */
   OPTMETAKEY,		/* interpret ALT-x as ESC-x */
   OPTAUXSORT,		/* (pseudo) using auxillary sort function */
@@ -270,6 +309,7 @@ enum
 /* Bit fields for ``Signals'' */
 #define S_INTERRUPT (1<<1)
 #define S_SIGWINCH  (1<<2)
+#define S_ALARM     (1<<3)
 
 typedef struct list_t
 {
@@ -278,7 +318,7 @@ typedef struct list_t
 } LIST;
 
 #define mutt_new_list() safe_calloc (1, sizeof (LIST))
-void mutt_add_to_list (LIST **, const char *);
+void mutt_add_to_list (LIST **, BUFFER *);
 void mutt_free_list (LIST **);
 int mutt_matches_ignore (const char *, LIST *);
 
@@ -292,6 +332,8 @@ typedef struct alias
   char *name;
   ADDRESS *addr;
   struct alias *next;
+  short tagged;
+  short num;
 } ALIAS;
 
 typedef struct envelope
@@ -371,6 +413,7 @@ typedef struct body
 				 * free()ing this structure
 				 */
   unsigned int tagged : 1;
+  unsigned int deleted : 1;	/* attachment marked for deletion */
 
 } BODY;
 
@@ -382,12 +425,24 @@ typedef struct header
   unsigned int tagged : 1;
   unsigned int deleted : 1;
   unsigned int changed : 1;
+  unsigned int attach_del : 1; /* has an attachment marked for deletion */
   unsigned int old : 1;
   unsigned int read : 1;
   unsigned int expired : 1; /* already expired? */
   unsigned int superseded : 1; /* got superseded? */
 
-  unsigned int pgp : 3;
+
+
+
+
+
+
+
+
+
+
+
+
   unsigned int replied : 1;
   unsigned int subject_changed : 1; /* used for threading */
   unsigned int display_subject : 1; /* used for threading */
@@ -403,6 +458,8 @@ typedef struct header
   unsigned int searched : 1;
   unsigned int matched : 1;
 
+  int pair; /* color-pair to use when displaying in the index */
+
   time_t date_sent;     /* time when the message was sent (UTC) */
   time_t received;      /* time when the message was placed in the mailbox */
   long offset;          /* where in the stream does this message begin? */
@@ -425,12 +482,35 @@ typedef struct header
 
 } HEADER;
 
+#include "mutt_regex.h"
+
+/* flag to mutt_pattern_comp() */
+#define M_FULL_MSG	1	/* enable body and header matching */
+
+typedef enum {
+  M_MATCH_FULL_ADDRESS = 1
+} pattern_exec_flag;
+
+typedef struct pattern_t
+{
+  short op;
+  short not;
+  int min;
+  int max;
+  struct pattern_t *next;
+  struct pattern_t *child;		/* arguments to logical op */
+  regex_t *rx;
+} pattern_t;
+
 typedef struct
 {
   char *path;
   FILE *fp;
   time_t mtime;
   off_t size;
+  off_t vsize;
+  char *pattern;                /* limit pattern string */
+  pattern_t *limit_pattern;     /* compiled limit pattern */
   HEADER **hdrs;
   HEADER *tree;			/* top of thread tree */
   HASH *id_hash;		/* hash table by msg id */
@@ -460,7 +540,6 @@ typedef struct
   unsigned int setgid : 1;
   unsigned int quiet : 1;	/* inhibit status messages? */
   unsigned int revsort : 1;	/* mailbox sorted in reverse? */
-
 } CONTEXT;
 
 typedef struct attachptr
@@ -485,6 +564,7 @@ typedef struct
 
 
 
+
 #define state_puts(x,y) fputs(x,(y)->fpout)
 #define state_putc(x,y) fputc(x,(y)->fpout)
 
