@@ -42,7 +42,7 @@
 typedef struct
 {
   char *pattern;	/* printable version */
-  void *rx; 		/* avoid including "mutt_regex.h" everywhere... */
+  regex_t *rx; 		/* compiled expression */
   int not;		/* do not match */
 } REGEXP;
 
