@@ -22,6 +22,11 @@
 #include "mime.h"
 #include "rfc2047.h"
 
+
+
+
+
+
 #include <string.h>
 #include <ctype.h>
 #include <sys/stat.h>
@@ -657,7 +662,7 @@ static time_t parse_date (char *s, HEADER *h)
     switch (count)
     {
       case 0: /* day of the month */
-	if (!isdigit (*t))
+	if (!isdigit ((unsigned char) *t))
 	  return (-1);
 	tm.tm_mday = atoi (t);
 	if (tm.tm_mday > 31)
@@ -699,7 +704,9 @@ static time_t parse_date (char *s, HEADER *h)
 
 	if (*ptz == '+' || *ptz == '-')
 	{
-	  if (ptz[1] && ptz[2] && ptz[3] && ptz[4])
+	  if (ptz[1] && ptz[2] && ptz[3] && ptz[4]
+	      && isdigit ((unsigned char) ptz[1]) && isdigit ((int) ptz[2])
+	      && isdigit ((unsigned char) ptz[3]) && isdigit ((int) ptz[4]))
 	  {
 	    zhours = (ptz[1] - '0') * 10 + (ptz[2] - '0');
 	    zminutes = (ptz[3] - '0') * 10 + (ptz[4] - '0');
@@ -800,6 +807,11 @@ void mutt_parse_mime_message (CONTEXT *ctx, HEADER *cur)
       cur->content->subtype = safe_strdup ("plain");
     }
     
+
+
+
+
+
     mx_close_message (&msg);
   }
 }
