@@ -17,11 +17,9 @@
  */ 
 
 #include "mutt.h"
-#include "mutt_regex.h"
 #include "mapping.h"
 #include "keymap.h"
 #include "mailbox.h"
-#include "pattern.h"
 #include "copy.h"
 
 #include <string.h>
@@ -35,21 +33,22 @@
 
 
 
-static const char *eat_regexp (pattern_t *pat, const char *s, char *, size_t);
-static const char *eat_date (pattern_t *pat, const char *s, char *, size_t);
-static const char *eat_range (pattern_t *pat, const char *s, char *, size_t);
+static int eat_regexp (pattern_t *pat, BUFFER *, BUFFER *);
+static int eat_date (pattern_t *pat, BUFFER *, BUFFER *);
+static int eat_range (pattern_t *pat, BUFFER *, BUFFER *);
 
 struct pattern_flags
 {
   int tag;	/* character used to represent this op */
   int op;	/* operation to perform */
   int class;
-  const char *(*eat_arg) (pattern_t *, const char *, char *, size_t);
+  int (*eat_arg) (pattern_t *, BUFFER *, BUFFER *);
 }
 Flags[] =
 {
   { 'A', M_ALL,			0,		NULL },
   { 'b', M_BODY,		M_FULL_MSG,	eat_regexp },
+  { 'B', M_WHOLE_MSG,		M_FULL_MSG,	eat_regexp },
   { 'c', M_CC,			0,		eat_regexp },
   { 'C', M_RECIPIENT,		0,		eat_regexp },
   { 'd', M_DATE,		0,		eat_date },
@@ -77,6 +76,7 @@ Flags[] =
   { 't', M_TO,			0,		eat_regexp },
   { 'U', M_UNREAD,		0,		NULL },
   { 'x', M_REFERENCE,		0,		eat_regexp },
+  { 'z', M_SIZE,		0,		eat_range },
   { 0 }
 };
 
@@ -105,7 +105,7 @@ int mutt_which_case (const char *s)
 {
   while (*s)
   {
-    if (isalpha (*s) && isupper (*s))
+    if (isalpha ((unsigned char) *s) && isupper ((int) *s))
       return 0; /* case-sensitive */
     s++;
   }
@@ -113,17 +113,18 @@ int mutt_which_case (const char *s)
 }
 
 static int
-msg_search (regex_t *rx, char *buf, size_t blen, int is_hdr, int msgno)
+msg_search (CONTEXT *ctx, regex_t *rx, char *buf, size_t blen, int op, int msgno)
 {
   char tempfile[_POSIX_PATH_MAX];
   MESSAGE *msg = NULL;
   STATE s;
   struct stat st;
   FILE *fp = NULL;
-  long lng;
+  long lng = 0;
   int match = 0;
+  HEADER *h = ctx->hdrs[msgno];
 
-  if ((msg = mx_open_message (Context, msgno)) != NULL)
+  if ((msg = mx_open_message (ctx, msgno)) != NULL)
   {
     if (option (OPTTHOROUGHSRC))
     {
@@ -137,13 +138,20 @@ msg_search (regex_t *rx, char *buf, size_t blen, int is_hdr, int msgno)
 	return (0);
       }
 
-      if (is_hdr)
-	mutt_copy_header (msg->fp, Context->hdrs[msgno], s.fpout, CH_FROM | CH_DECODE, NULL);
-      else
+      if (op != M_BODY)
+	mutt_copy_header (msg->fp, h, s.fpout, CH_FROM | CH_DECODE, NULL);
+
+      if (op != M_HEADER)
       {
-	mutt_parse_mime_message (Context, Context->hdrs[msgno]);
-	fseek (msg->fp, Context->hdrs[msgno]->offset, 0);
-	mutt_body_handler (Context->hdrs[msgno]->content, &s);
+	mutt_parse_mime_message (ctx, h);
+
+
+
+
+
+
+	fseek (msg->fp, h->offset, 0);
+	mutt_body_handler (h->content, &s);
       }
 
       fp = s.fpout;
@@ -156,15 +164,16 @@ msg_search (regex_t *rx, char *buf, size_t blen, int is_hdr, int msgno)
     {
       /* raw header / body */
       fp = msg->fp;
-      if (is_hdr)
+      if (op != M_BODY)
       {
-	fseek (fp, Context->hdrs[msgno]->offset, 0);
-	lng = Context->hdrs[msgno]->content->offset - Context->hdrs[msgno]->offset;
+	fseek (fp, h->offset, 0);
+	lng = h->content->offset - h->offset;
       }
-      else
+      if (op != M_HEADER)
       {
-	fseek (msg->fp, Context->hdrs[msgno]->content->offset, 0);
-	lng = Context->hdrs[msgno]->content->length;
+	if (op == M_BODY)
+	  fseek (fp, h->content->offset, 0);
+	lng += h->content->length;
       }
     }
 
@@ -193,67 +202,98 @@ msg_search (regex_t *rx, char *buf, size_t blen, int is_hdr, int msgno)
   return match;
 }
 
-static const char *
-eat_regexp (pattern_t *pat, const char *s, char *err, size_t errlen)
+int eat_regexp (pattern_t *pat, BUFFER *s, BUFFER *err)
 {
-  char buf[SHORT_STRING];
-  const char *ps;
+  BUFFER buf;
   int r;
 
-  ps = mutt_extract_token (buf, sizeof (buf), s, NULL, 0,
-			   M_PATTERN | M_COMMENT | M_NULL);
-  
-  pat->rx = safe_malloc (sizeof (regex_t));
-  if ((r = REGCOMP (pat->rx, buf, REG_NEWLINE | REG_NOSUB | mutt_which_case (buf))) != 0)
+  memset (&buf, 0, sizeof (buf));
+  if (mutt_extract_token (&buf, s, M_TOKEN_PATTERN | M_TOKEN_COMMENT) != 0 ||
+      !buf.data)
   {
-    regerror (r, pat->rx, err, errlen);
+    snprintf (err->data, err->dsize, "Error in expression: %s", s->dptr);
+    return (-1);
+  }
+  pat->rx = safe_malloc (sizeof (regex_t));
+  r = REGCOMP (pat->rx, buf.data, REG_NEWLINE | REG_NOSUB | mutt_which_case (buf.data));
+  FREE (&buf.data);
+  if (r)
+  {
+    regerror (r, pat->rx, err->data, err->dsize);
     regfree (pat->rx);
     safe_free ((void **) &pat->rx);
-    return NULL;
+    return (-1);
   }
-
-  return ps;
+  return 0;
 }
 
-static const char *
-eat_range (pattern_t *pat, const char *s, char *err, size_t errlen)
+int eat_range (pattern_t *pat, BUFFER *s, BUFFER *err)
 {
-  const char *p;
   char *tmp;
 
-  if (*s != '-')
+  if ((*s->dptr != '-') && (*s->dptr != '<'))
   {
     /* range minimum */
-    pat->min = strtol (s, &tmp, 0);
-    p = tmp;
+    if (*s->dptr == '>')
+    {
+      pat->max = M_MAXRANGE;
+      pat->min = strtol (s->dptr + 1, &tmp, 0);
+    }
+    else
+      pat->min = strtol (s->dptr, &tmp, 0);
+    if (toupper (*tmp) == 'K') /* is there a prefix? */
+    {
+      pat->min *= 1024;
+      tmp++;
+    }
+    else if (toupper (*tmp) == 'M')
+    {
+      pat->min *= 1048576;
+      tmp++;
+    }
+    if (*s->dptr == '>')
+    {
+      s->dptr = tmp;
+      return 0;
+    }
+    if (*tmp != '-')
+    {
+      /* exact value */
+      pat->max = pat->min;
+      s->dptr = tmp;
+      return 0;
+    }
+    tmp++;
   }
   else
   {
-    s++;
-    p = s;
+    s->dptr++;
+    tmp = s->dptr;
   }
-
-  if (*p != '-')
+  
+  if (isdigit ((unsigned char) *tmp))
   {
-    pat->max = pat->min;
-    return p;
-  }
-  p++;
-
-  if (isdigit (*p))
-  {
-    /* range max */
-    pat->max = strtol (p, &tmp, 0);
-    p = tmp;
+    /* range maximum */
+    pat->max = strtol (tmp, &tmp, 0);
+    if (toupper (*tmp) == 'K')
+    {
+      pat->max *= 1024;
+      tmp++;
+    }
+    else if (toupper (*tmp) == 'M')
+    {
+      pat->max *= 1048576;
+      tmp++;
+    }
   }
   else
     pat->max = M_MAXRANGE;
 
-  return p;
+  s->dptr = tmp;
+  return 0;
 }
 
-static const char *
-getDate (const char *s, struct tm *t, char *err, size_t errlen)
+static const char *getDate (const char *s, struct tm *t, BUFFER *err)
 {
   char *p;
   time_t now = time (NULL);
@@ -262,7 +302,7 @@ getDate (const char *s, struct tm *t, char *err, size_t errlen)
   t->tm_mday = strtol (s, &p, 0);
   if (t->tm_mday < 1 || t->tm_mday > 31)
   {
-    snprintf (err, errlen, "Invalid day of month: %s", s);
+    snprintf (err->data, err->dsize, "Invalid day of month: %s", s);
     return NULL;
   }
   if (*p != '/')
@@ -276,7 +316,7 @@ getDate (const char *s, struct tm *t, char *err, size_t errlen)
   t->tm_mon = strtol (p, &p, 0) - 1;
   if (t->tm_mon < 0 || t->tm_mon > 11)
   {
-    snprintf (err, errlen, "Invalid month: %s", p);
+    snprintf (err->data, err->dsize, "Invalid month: %s", p);
     return NULL;
   }
   if (*p != '/')
@@ -321,14 +361,18 @@ static const char *get_offset (struct tm *tm, const char *s)
   return (ps + 1);
 }
 
-static const char *
-eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
+static int eat_date (pattern_t *pat, BUFFER *s, BUFFER *err)
 {
-  char buffer[SHORT_STRING];
+  BUFFER buffer;
   struct tm min, max;
 
-  s = mutt_extract_token (buffer, sizeof (buffer), s, NULL, 0,
-			  M_NULL | M_COMMENT | M_PATTERN);
+  memset (&buffer, 0, sizeof (buffer));
+  if (mutt_extract_token (&buffer, s, M_TOKEN_COMMENT | M_TOKEN_PATTERN) != 0
+      || !buffer.data)
+  {
+    strfcpy (err->data, "error in expression", err->dsize);
+    return (-1);
+  }
 
   memset (&min, 0, sizeof (min));
   /* the `0' time is Jan 1, 1970 UTC, so in order to prevent a negative time
@@ -349,7 +393,7 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
   max.tm_min = 59;
   max.tm_sec = 59;
 
-  if (strchr ("<>=", buffer[0]))
+  if (strchr ("<>=", buffer.data[0]))
   {
     /* offset from current time
        <3d	less than three days ago
@@ -359,7 +403,7 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
     struct tm *tm = localtime (&now);
     int exact = 0;
 
-    if (buffer[0] == '<')
+    if (buffer.data[0] == '<')
     {
       memcpy (&min, tm, sizeof (min));
       tm = &min;
@@ -369,13 +413,13 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
       memcpy (&max, tm, sizeof (max));
       tm = &max;
 
-      if (buffer[0] == '=')
+      if (buffer.data[0] == '=')
 	exact++;
     }
     tm->tm_hour = 23;
     tm->tm_min = max.tm_sec = 59;
 
-    get_offset (tm, buffer + 1);
+    get_offset (tm, buffer.data + 1);
 
     if (exact)
     {
@@ -386,13 +430,16 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
   }
   else
   {
-    const char *pc = buffer;
+    const char *pc = buffer.data;
 
     if (*pc != '-')
     {
       /* mininum date specified */
-      if ((pc = getDate (pc, &min, err, errlen)) == NULL)
-	return NULL;
+      if ((pc = getDate (pc, &min, err)) == NULL)
+      {
+	FREE (&buffer.data);
+	return (-1);
+      }
     }
 
     if (*pc && *pc == '-')
@@ -401,8 +448,11 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
       pc++; /* skip the `-' */
       SKIPWS (pc);
       if (*pc)
-	if (!getDate (pc, &max, err, errlen))
-	  return NULL;
+	if (!getDate (pc, &max, err))
+	{
+	  FREE (&buffer.data);
+	  return (-1);
+	}
     }
     else
     {
@@ -416,7 +466,9 @@ eat_date (pattern_t *pat, const char *s, char *err, size_t errlen)
   pat->min = mutt_mktime (&min, 1);
   pat->max = mutt_mktime (&max, 1);
 
-  return s;
+  FREE (&buffer.data);
+
+  return 0;
 }
 
 static struct pattern_flags *lookup_tag (char tag)
@@ -429,7 +481,7 @@ static struct pattern_flags *lookup_tag (char tag)
   return NULL;
 }
 
-static const char *find_matching_paren (const char *s)
+static /* const */ char *find_matching_paren (/* const */ char *s)
 {
   int level = 1;
 
@@ -467,10 +519,7 @@ void mutt_pattern_free (pattern_t **pat)
   }
 }
 
-pattern_t *mutt_pattern_comp (const char *s,
-			      int flags,
-			      char *err,
-			      size_t errlen)
+pattern_t *mutt_pattern_comp (/* const */ char *s, int flags, BUFFER *err)
 {
   pattern_t *curlist = NULL;
   pattern_t *tmp;
@@ -479,16 +528,21 @@ pattern_t *mutt_pattern_comp (const char *s,
   int or = 0;
   int implicit = 1;	/* used to detect logical AND operator */
   struct pattern_flags *entry;
-  const char *p;
-  char buf[STRING];
+  char *p;
+  char *buf;
+  BUFFER ps;
 
-  while (*s)
+  memset (&ps, 0, sizeof (ps));
+  ps.dptr = s;
+  ps.dsize = strlen (s);
+
+  while (*ps.dptr)
   {
-    SKIPWS (s);
-    switch (*s)
+    SKIPWS (ps.dptr);
+    switch (*ps.dptr)
     {
       case '!':
-	s++;
+	ps.dptr++;
 	not = !not;
 	break;
       case '|':
@@ -496,7 +550,7 @@ pattern_t *mutt_pattern_comp (const char *s,
 	{
 	  if (!curlist)
 	  {
-	    snprintf (err, errlen, "error in pattern at: %s", s);
+	    snprintf (err->data, err->dsize, "error in pattern at: %s", ps.dptr);
 	    return NULL;
 	  }
 	  if (curlist->next)
@@ -512,7 +566,7 @@ pattern_t *mutt_pattern_comp (const char *s,
 
 	  or = 1;
 	}
-	s++;
+	ps.dptr++;
 	implicit = 0;
 	not = 0;
 	break;
@@ -538,33 +592,33 @@ pattern_t *mutt_pattern_comp (const char *s,
 	  curlist = tmp;
 	last = tmp;
 
-	s++; /* move past the ~ */
-	if ((entry = lookup_tag (*s)) == NULL)
+	ps.dptr++; /* move past the ~ */
+	if ((entry = lookup_tag (*ps.dptr)) == NULL)
 	{
-	  snprintf (err, errlen, "%c: invalid command", *s);
+	  snprintf (err->data, err->dsize, "%c: invalid command", *ps.dptr);
 	  mutt_pattern_free (&curlist);
 	  return NULL;
 	}
 	if (entry->class && (flags & entry->class) == 0)
 	{
-	  snprintf (err, errlen, "%c: not supported in this mode", *s);
+	  snprintf (err->data, err->dsize, "%c: not supported in this mode", *ps.dptr);
 	  mutt_pattern_free (&curlist);
 	  return NULL;
 	}
 	tmp->op = entry->op;
 
-	s++; /* eat the operator and any optional whitespace */
-	SKIPWS (s);
+	ps.dptr++; /* eat the operator and any optional whitespace */
+	SKIPWS (ps.dptr);
 
 	if (entry->eat_arg)
 	{
-	  if (!*s)
+	  if (!*ps.dptr)
 	  {
-	    snprintf (err, errlen, "missing parameter");
+	    snprintf (err->data, err->dsize, "missing parameter");
 	    mutt_pattern_free (&curlist);
 	    return NULL;
 	  }
-	  if ((s = entry->eat_arg (tmp, s, err, errlen)) == NULL)
+	  if (entry->eat_arg (tmp, &ps, err) == -1)
 	  {
 	    mutt_pattern_free (&curlist);
 	    return NULL;
@@ -573,20 +627,22 @@ pattern_t *mutt_pattern_comp (const char *s,
 	implicit = 1;
 	break;
       case '(':
-	p = find_matching_paren (s + 1);
+	p = find_matching_paren (ps.dptr + 1);
 	if (*p != ')')
 	{
-	  snprintf (err, errlen, "mismatched parenthesis: %s", s);
+	  snprintf (err->data, err->dsize, "mismatched parenthesis: %s", ps.dptr);
 	  mutt_pattern_free (&curlist);
 	  return NULL;
 	}
-	mutt_substrcpy (buf, s + 1, p, sizeof (buf));
 	/* compile the sub-expression */
-	if ((tmp = mutt_pattern_comp (buf, flags, err, errlen)) == NULL)
+	buf = mutt_substrdup (ps.dptr + 1, p);
+	if ((tmp = mutt_pattern_comp (buf, flags, err)) == NULL)
 	{
+	  FREE (&buf);
 	  mutt_pattern_free (&curlist);
 	  return NULL;
 	}
+	FREE (&buf);
 	if (last)
 	  last->next = tmp;
 	else
@@ -594,17 +650,17 @@ pattern_t *mutt_pattern_comp (const char *s,
 	last = tmp;
 	tmp->not = not;
 	not = 0;
-	s = p + 1; /* restore location */
+	ps.dptr = p + 1; /* restore location */
 	break;
       default:
-	snprintf (err, errlen, "error in pattern at: %s", s);
+	snprintf (err->data, err->dsize, "error in pattern at: %s", ps.dptr);
 	mutt_pattern_free (&curlist);
 	return NULL;
     }
   }
   if (!curlist)
   {
-    strfcpy (err, "empty pattern", errlen);
+    strfcpy (err->data, "empty pattern", err->dsize);
     return NULL;
   }
   if (curlist->next)
@@ -614,7 +670,7 @@ pattern_t *mutt_pattern_comp (const char *s,
     tmp->child = curlist;
     curlist = tmp;
   }
-  return curlist;
+  return (curlist);
 }
 
 static int
@@ -705,9 +761,9 @@ mutt_pattern_exec (struct pattern_t *pat, pattern_exec_flag flags, CONTEXT *ctx,
     case M_DATE_RECEIVED:
       return (pat->not ^ (h->received >= pat->min && h->received <= pat->max));
     case M_BODY:
-      return (pat->not ^ msg_search (pat->rx, buf, sizeof (buf), 0, h->msgno));
     case M_HEADER:
-      return (pat->not ^ msg_search (pat->rx, buf, sizeof (buf), 1, h->msgno));
+    case M_WHOLE_MSG:
+      return (pat->not ^ msg_search (ctx, pat->rx, buf, sizeof (buf), pat->op, h->msgno));
     case M_SENDER:
       return (pat->not ^ match_adrlist (pat->rx, flags & M_MATCH_FULL_ADDRESS, h->env->sender));
     case M_FROM:
@@ -723,6 +779,8 @@ mutt_pattern_exec (struct pattern_t *pat, pattern_exec_flag flags, CONTEXT *ctx,
     case M_SCORE:
       return (pat->not ^ (h->score >= pat->min && (pat->max == M_MAXRANGE ||
 						   h->score <= pat->max)));
+    case M_SIZE:
+      return (pat->not ^ (h->content->length > pat->min && (pat->max == M_MAXRANGE || h->content->length < pat->max)));
     case M_REFERENCE:
       return (pat->not ^ match_reference (pat->rx, h->env->references));
     case M_ADDRESS:
@@ -798,20 +856,24 @@ void mutt_check_simple (char *s, size_t len, const char *simple)
 int mutt_pattern_func (int op, char *prompt, HEADER *hdr)
 {
   pattern_t *pat;
-  char buf[STRING] = "";
-  char error[STRING];
+  char buf[LONG_STRING] = "", *simple, error[STRING];
+  BUFFER err;
   int i;
 
   if (mutt_get_field (prompt, buf, sizeof (buf), 0) != 0 || !buf[0])
     return (-1);
 
   mutt_message ("Compiling search pattern...");
+  
+  simple = safe_strdup (buf);
+  mutt_check_simple (buf, sizeof (buf), NONULL (SimpleSearch));
 
-  mutt_check_simple (buf, sizeof (buf), SimpleSearch);
-
-  if ((pat = mutt_pattern_comp (buf, M_FULL_MSG, error, sizeof (error))) == NULL)
+  err.data = error;
+  err.dsize = sizeof (error);
+  if ((pat = mutt_pattern_comp (buf, M_FULL_MSG, &err)) == NULL)
   {
-    mutt_error ("%s", error);
+    FREE (&simple);
+    mutt_error ("%s", err.data);
     return (-1);
   }
 
@@ -822,8 +884,11 @@ int mutt_pattern_func (int op, char *prompt, HEADER *hdr)
     for (i = 0; i < Context->msgcount; i++)
       Context->hdrs[i]->virtual = -1;
     Context->vcount = 0;
+    Context->vsize = 0;
   }
 
+#define this_body Context->hdrs[i]->content
+
   for (i = 0; i < Context->msgcount; i++)
     if (mutt_pattern_exec (pat, M_MATCH_FULL_ADDRESS, Context, Context->hdrs[i]))
     {
@@ -845,28 +910,41 @@ int mutt_pattern_func (int op, char *prompt, HEADER *hdr)
 	  Context->hdrs[i]->virtual = Context->vcount;
 	  Context->v2r[Context->vcount] = i;
 	  Context->vcount++;
+	  Context->vsize+=this_body->length + this_body->offset -
+	                  this_body->hdr_offset;
 	  break;
       }
     }
+#undef this_body
 
   mutt_clear_error ();
 
-  if (op == M_LIMIT && !Context->vcount)
+  if (op == M_LIMIT)
   {
-    mutt_error ("No messages matched criteria.");
-    /* restore full display */
-    Context->vcount = 0;
-    for (i = 0; i < Context->msgcount; i++)
+    safe_free ((void **) &Context->pattern);
+    if (Context->limit_pattern) 
+      mutt_pattern_free (&Context->limit_pattern);
+    if (!Context->vcount)
     {
-      Context->hdrs[i]->virtual = i;
-      Context->v2r[i] = i;
+      mutt_error ("No messages matched criteria.");
+      /* restore full display */
+      for (i = 0; i < Context->msgcount; i++)
+      {
+	Context->hdrs[i]->virtual = i;
+	Context->v2r[i] = i;
+      }
+
+      Context->vcount = Context->msgcount;
+    }
+    else if (strncmp (buf, "~A", 2) != 0)
+    {
+      Context->pattern = simple;
+      simple = NULL; /* don't clobber it */
+      Context->limit_pattern = mutt_pattern_comp (buf, M_FULL_MSG, &err);
     }
-
-    Context->vcount = Context->msgcount;
   }
-
+  FREE (&simple);
   mutt_pattern_free (&pat);
-
   return 0;
 }
 
@@ -876,6 +954,7 @@ int mutt_search_command (int cur, int op)
   char buf[STRING];
   char temp[LONG_STRING];
   char error[STRING];
+  BUFFER err;
   int incr;
   HEADER *h;
   
@@ -894,7 +973,7 @@ int mutt_search_command (int cur, int op)
     /* compare the *expanded* version of the search pattern in case 
        $simple_search has changed while we were searching */
     strfcpy (temp, buf, sizeof (temp));
-    mutt_check_simple (temp, sizeof (temp), SimpleSearch);
+    mutt_check_simple (temp, sizeof (temp), NONULL (SimpleSearch));
 
     if (!SearchPattern || strcmp (temp, LastSearchExpn))
     {
@@ -902,7 +981,9 @@ int mutt_search_command (int cur, int op)
       strfcpy (LastSearch, buf, sizeof (LastSearch));
       mutt_message ("Compiling search pattern...");
       mutt_pattern_free (&SearchPattern);
-      if ((SearchPattern = mutt_pattern_comp (temp, M_FULL_MSG, error, sizeof (error))) == NULL)
+      err.data = error;
+      err.dsize = sizeof (error);
+      if ((SearchPattern = mutt_pattern_comp (temp, M_FULL_MSG, &err)) == NULL)
       {
 	mutt_error ("%s", error);
 	return (-1);
@@ -932,12 +1013,24 @@ int mutt_search_command (int cur, int op)
     if (i > Context->vcount - 1)
     {
       i = 0;
-      mutt_message ("Search wrapped to top.");
+      if (option (OPTWRAPSEARCH))
+        mutt_message ("Search wrapped to top.");
+      else 
+      {
+        mutt_message ("Search hit bottom without finding match");
+	return (-1);
+      }
     }
     else if (i < 0)
     {
       i = Context->vcount - 1;
-      mutt_message ("Search wrapped to bottom.");
+      if (option (OPTWRAPSEARCH))
+        mutt_message ("Search wrapped to bottom.");
+      else 
+      {
+        mutt_message ("Search hit top without finding match");
+	return (-1);
+      }
     }
 
     h = Context->hdrs[Context->v2r[i]];
@@ -955,6 +1048,13 @@ int mutt_search_command (int cur, int op)
 	return i;
     }
 
+    if (Signals & S_INTERRUPT)
+    {
+      mutt_error ("Search interrupted.");
+      Signals &= ~S_INTERRUPT;
+      return (-1);
+    }
+
     i += incr;
   }
 
