@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-1998 Michael R. Elkins <me@cs.hmc.edu>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -61,9 +61,9 @@ void mutt_update_tree (ATTACHPTR **idx, short idxlen)
       if (idx[x]->level)
       {
 	s = buf + 2 * (idx[x]->level - 1);
-	*s++ = (idx[x]->content->next) ? '\003' : '\001';
-	*s++ = '\004';
-	*s++ = '\007';
+	*s++ = (idx[x]->content->next) ? M_TREE_LTEE : M_TREE_LLCORNER;
+	*s++ = M_TREE_HLINE;
+	*s++ = M_TREE_RARROW;
       }
       else
 	s = buf;
@@ -142,12 +142,14 @@ void attach_entry (char *b, size_t blen, MUTTMENU *menu, int num)
   s[0] = 0;
   if (m->type == TYPEMESSAGE && (!strcasecmp ("rfc822", m->subtype) ||
       !strcasecmp ("news", m->subtype)) && MsgFmt[0])
-    _mutt_make_string (s, sizeof (s), MsgFmt, m->hdr, M_FORCESUBJ | M_MAKEPRINT);
+    _mutt_make_string (s, sizeof (s), MsgFmt, NULL, m->hdr,
+		       M_FORMAT_FORCESUBJ | M_FORMAT_MAKEPRINT);
 
   mutt_pretty_size (size, sizeof (size), m->length);
   snprintf (t, sizeof (t), "[%.7s/%.10s, %.6s, %s]",
 	    TYPE (m->type), m->subtype, ENCODING (m->encoding), size);
-  snprintf (b, blen, " %c %2d %-34.34s %s%s",
+  snprintf (b, blen, " %c%c %2d %-34.34s %s%s",
+	    m->deleted ? 'D' : ' ',
 	    m->tagged ? '*' : ' ',
 	    num + 1,
 	    t,
@@ -203,8 +205,10 @@ mutt_query_pipe_attachment (char *command, FILE *fp, BODY *body, int filter)
     snprintf (warning, sizeof (warning),
 	      "WARNING!  You are about to overwrite %s, continue?",
 	      body->filename);
-    if (mutt_yesorno (warning, M_NO) != M_YES)
+    if (mutt_yesorno (warning, M_NO) != M_YES) {
+      CLEARLINE (LINES-1);
       return;
+    }
     mutt_mktemp (tfile);
   }
   else
@@ -484,6 +488,12 @@ mutt_attach_display_loop (MUTTMENU *menu, int op, FILE *fp, ATTACHPTR **idx)
 
 void mutt_view_attachments (HEADER *hdr)
 {
+
+
+
+
+
+
   char helpstr[SHORT_STRING];
   MUTTMENU *menu;
   BODY *cur;
@@ -501,6 +511,21 @@ void mutt_view_attachments (HEADER *hdr)
   if ((msg = mx_open_message (Context, hdr->msgno)) == NULL)
     return;
 
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
   {
     fp = msg->fp;
     cur = hdr->content;
@@ -537,7 +562,11 @@ void mutt_view_attachments (HEADER *hdr)
 	menu->redraw = REDRAW_FULL;
 	break;
 
+
+
       
+
+
       case OP_PRINT:
 	mutt_print_attachment_list (fp, menu->tagprefix, menu->tagprefix ? cur : idx[menu->current]->content);
 	break;
@@ -548,8 +577,79 @@ void mutt_view_attachments (HEADER *hdr)
 
       case OP_SAVE:
 	mutt_save_attachment_list (fp, menu->tagprefix, menu->tagprefix ? cur : idx[menu->current]->content);
+	if (option (OPTRESOLVE) && menu->current < menu->max - 1)
+	{
+	  menu->current++;
+	  menu->redraw = REDRAW_MOTION_RESYNCH;
+	}
+	else
+	  menu->redraw = REDRAW_CURRENT;
 	break;
 
+      case OP_DELETE:
+
+       if (menu->max == 1)
+       {
+         mutt_message ("Only deletion of multipart attachments is supported.");
+       }
+       else
+       {
+         {
+	   if (!menu->tagprefix)
+	   {
+	     idx[menu->current]->content->deleted = 1;
+	     if (option (OPTRESOLVE) && menu->current < menu->max - 1)
+	     {
+	       menu->current++;
+	       menu->redraw = REDRAW_MOTION_RESYNCH;
+	     }
+	     else
+	       menu->redraw = REDRAW_CURRENT;
+	   }
+	   else
+	   {
+	     int x;
+
+	     for (x = 0; x < menu->max; x++)
+	     {
+	       if (idx[x]->content->tagged)
+	       {
+		 idx[x]->content->deleted = 1;
+		 menu->redraw = REDRAW_INDEX;
+	       }
+	     }
+	   }
+         }
+       }
+       break;
+
+      case OP_UNDELETE:
+       if (!menu->tagprefix)
+       {
+	 idx[menu->current]->content->deleted = 0;
+	 if (option (OPTRESOLVE) && menu->current < menu->max - 1)
+	 {
+	   menu->current++;
+	   menu->redraw = REDRAW_MOTION_RESYNCH;
+	 }
+	 else
+	   menu->redraw = REDRAW_CURRENT;
+       }
+       else
+       {
+	 int x;
+
+	 for (x = 0; x < menu->max; x++)
+	 {
+	   if (idx[x]->content->tagged)
+	   {
+	     idx[x]->content->deleted = 0;
+	     menu->redraw = REDRAW_INDEX;
+	   }
+	 }
+       }
+       break;
+
       case OP_BOUNCE_MESSAGE:
 	query_bounce_attachment (menu->tagprefix, menu->tagprefix ? cur : idx[menu->current]->content, hdr);
 	break;
@@ -558,6 +658,12 @@ void mutt_view_attachments (HEADER *hdr)
       case OP_GROUP_REPLY:
       case OP_LIST_REPLY:
       case OP_FORWARD_MESSAGE:
+
+
+
+
+
+
 	if (op == OP_FORWARD_MESSAGE)
 	  flags = SENDFORWARD;
 	else
@@ -573,15 +679,32 @@ void mutt_view_attachments (HEADER *hdr)
 
       case OP_EXIT:
 	mx_close_message (&msg);
+	hdr->attach_del = 0;
 	while (idxlen-- > 0)
 	{
+	  if (idx[idxlen]->content->deleted)
+	    hdr->attach_del = 1;
 	  safe_free ((void **) &idx[idxlen]->tree);
 	  safe_free ((void **) &idx[idxlen]);
 	}
+	if (hdr->attach_del)
+	  hdr->changed = 1;
 	safe_free ((void **) &idx);
 	idxmax = 0;
 
 
+
+
+
+
+
+
+
+
+
+
+
+
 	mutt_menuDestroy  (&menu);
 	return;
     }
