@@ -33,41 +33,20 @@
 #include <sys/wait.h>
 #include <dirent.h>
 
+extern char RFC822Specials[];
 
 
 
 
 
-FILE *mutt_open_read (const char *path, pid_t *thepid)
-{
-  FILE *f;
-  int len = strlen (path);
 
-  if (path[len-1] == '|')
-  {
-    /* read from a pipe */
-
-    char *s = safe_strdup (path);
-
-    s[len-1] = 0;
-    endwin ();
-    *thepid = mutt_create_filter (s, NULL, &f, NULL);
-    free (s);
-  }
-  else
-  {
-    f = fopen (path, "r");
-    *thepid = -1;
-  }
-  return (f);
-}
 
 static void append_signature (ENVELOPE *env, FILE *f)
 {
   FILE *tmpfp;
   pid_t thepid;
 
-  if ((tmpfp = mutt_open_read (Signature, &thepid)))
+  if (Signature && (tmpfp = mutt_open_read (Signature, &thepid)))
   {
     if (option (OPTSIGDASHES))
       fputs ("\n-- \n", f);
@@ -107,12 +86,10 @@ static ADDRESS *mutt_remove_xrefs (ADDRESS *a, ADDRESS *b)
   top = b;
   while (b)
   {
-    p = a;
-    while (p)
+    for (p = a; p; p = p->next)
     {
       if (mutt_addrcmp (p, b))
 	break;
-      p = p->next;
     }
     if (p)
     {
@@ -201,39 +178,30 @@ static ADDRESS *find_mailing_lists (ADDRESS *t, ADDRESS *c)
   return top;
 }
 
-static int edit_envelope (ENVELOPE *en)
+static int edit_address (ADDRESS **a, /* const */ char *field)
 {
   char buf[HUGE_STRING];
 
   buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), en->to);
-  if (mutt_get_field ("To: ", buf, sizeof (buf), M_ALIAS) != 0 || !buf[0])
+  rfc822_write_address (buf, sizeof (buf), *a);
+  if (mutt_get_field (field, buf, sizeof (buf), M_ALIAS) != 0)
     return (-1);
-  rfc822_free_address (&en->to);
-  en->to = mutt_parse_adrlist (en->to, buf);
-  en->to = mutt_expand_aliases (en->to);
+  rfc822_free_address (a);
+  *a = mutt_expand_aliases (mutt_parse_adrlist (NULL, buf));
+  return 0;
+}
 
-  if (option (OPTASKCC))
-  {
-    buf[0] = 0;
-    rfc822_write_address (buf, sizeof (buf), en->cc);
-    if (mutt_get_field ("Cc: ", buf, sizeof (buf), M_ALIAS) != 0)
-      return (-1);
-    rfc822_free_address (&en->cc);
-    en->cc = mutt_parse_adrlist (en->cc, buf);
-    en->cc = mutt_expand_aliases (en->cc);
-  }
+static int edit_envelope (ENVELOPE *en)
+{
+  char buf[HUGE_STRING];
+  LIST *uh = UserHeader;
 
-  if (option (OPTASKBCC))
-  {
-    buf[0] = 0;
-    rfc822_write_address (buf, sizeof (buf), en->bcc);
-    if (mutt_get_field ("Bcc: ", buf, sizeof (buf), M_ALIAS) != 0)
-      return (-1);
-    rfc822_free_address (&en->bcc);
-    en->bcc = mutt_parse_adrlist (en->bcc, buf);
-    en->bcc = mutt_expand_aliases (en->bcc);
-  }
+  if (edit_address (&en->to, "To: ") == -1 || en->to == NULL)
+    return (-1);
+  if (option (OPTASKCC) && edit_address (&en->cc, "Cc: ") == -1)
+    return (-1);
+  if (option (OPTASKBCC) && edit_address (&en->bcc, "Bcc: ") == -1)
+    return (-1);
 
   if (en->subject)
   {
@@ -243,13 +211,27 @@ static int edit_envelope (ENVELOPE *en)
       strfcpy (buf, en->subject, sizeof (buf));
   }
   else
+  {
+    char *p;
+
     buf[0] = 0;
+    for (; uh; uh = uh->next)
+    {
+      if (strncasecmp ("subject:", uh->data, 8) == 0)
+      {
+	p = uh->data + 8;
+	SKIPWS (p);
+	strncpy (buf, p, sizeof (buf));
+      }
+    }
+  }
+  
   if (mutt_get_field ("Subject: ", buf, sizeof (buf), 0) != 0 ||
       (!buf[0] && query_quadoption (OPT_SUBJECT, "No subject, abort?") != 0))
-    {
-      mutt_message ("No subject, aborting.");
-      return (-1);
-    }
+  {
+    mutt_message ("No subject, aborting.");
+    return (-1);
+  }
   safe_free ((void **) &en->subject);
   en->subject = safe_strdup (buf);
 
@@ -284,9 +266,7 @@ static void process_user_header (ENVELOPE *env)
   {
     if (strncasecmp ("from:", uh->data, 5) == 0)
     {
-      /* User has specified a default From: address.  Get rid of the
-       * default address.
-       */
+      /* User has specified a default From: address.  Remove default address */
       rfc822_free_address (&env->from);
       env->from = rfc822_parse_adrlist (env->from, uh->data + 5);
     }
@@ -297,7 +277,8 @@ static void process_user_header (ENVELOPE *env)
     }
     else if (strncasecmp ("to:", uh->data, 3) != 0 &&
 	     strncasecmp ("cc:", uh->data, 3) != 0 &&
-	     strncasecmp ("bcc:", uh->data, 4) != 0)
+	     strncasecmp ("bcc:", uh->data, 4) != 0 &&
+	     strncasecmp ("subject:", uh->data, 8) != 0)
     {
       if (last)
       {
@@ -369,9 +350,9 @@ static int include_reply (CONTEXT *ctx, HEADER *cur, FILE *out)
 
 
 
-  if (Attribution[0])
+  if (Attribution)
   {
-    mutt_make_string (buffer, sizeof (buffer), Attribution, cur);
+    mutt_make_string (buffer, sizeof (buffer), Attribution, ctx, cur);
     fputs (buffer, out);
     fputc ('\n', out);
   }
@@ -379,9 +360,9 @@ static int include_reply (CONTEXT *ctx, HEADER *cur, FILE *out)
     flags |= M_CM_NOHEADER;
   mutt_parse_mime_message (ctx, cur);
   mutt_copy_message (out, ctx, cur, flags, CH_DECODE);
-  if (PostIndentString[0])
+  if (PostIndentString)
   {
-    mutt_make_string (buffer, sizeof (buffer), PostIndentString, cur);
+    mutt_make_string (buffer, sizeof (buffer), PostIndentString, ctx, cur);
     fputs (buffer, out);
     fputc ('\n', out);
   }
@@ -412,32 +393,14 @@ static BODY *make_forward (CONTEXT *ctx, HEADER *hdr)
 
   mutt_parse_mime_message (ctx, hdr);
   mutt_copy_message (fpout, ctx, hdr,
-		     option (OPTFORWDECODE) ? M_CM_DECODE : 0,
-		     CH_XMIT | (option (OPTFORWDECODE) ? (CH_MIME | CH_TXTPLAIN ) : 0));
+		     option (OPTMIMEFORWDECODE) ? M_CM_DECODE : 0,
+		     CH_XMIT | (option (OPTMIMEFORWDECODE) ? (CH_MIME | CH_TXTPLAIN ) : 0));
   
   fclose (fpout);
   mutt_update_encoding (body);
   return (body);
 }
 
-/* append list 'b' to list 'a' and return the last element in the new list */
-static ADDRESS *rfc822_append (ADDRESS **a, ADDRESS *b)
-{
-  ADDRESS *tmp = *a;
-
-  while (tmp && tmp->next)
-    tmp = tmp->next;
-  if (!b)
-    return tmp;
-  if (tmp)
-    tmp->next = rfc822_cpy_adr (b);
-  else
-    tmp = *a = rfc822_cpy_adr (b);
-  while (tmp && tmp->next)
-    tmp = tmp->next;
-  return tmp;
-}
-
 static int default_to (ADDRESS **to, ENVELOPE *env, int group)
 {
   char prompt[STRING];
@@ -463,7 +426,8 @@ static int default_to (ADDRESS **to, ENVELOPE *env, int group)
   {
     if (option (OPTIGNORELISTREPLYTO) &&
 	mutt_is_mail_list (env->reply_to) &&
-	mutt_addrsrc (env->reply_to, env->to))
+	(mutt_addrsrc (env->reply_to, env->to) ||
+	mutt_addrsrc (env->reply_to, env->cc)))
     {
       /* If the Reply-To: address is a mailing list, assume that it was
        * put there by the mailing list, and use the From: address
@@ -594,19 +558,22 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
       env->subject = safe_strdup ("Re: your mail");
 
     /* add the In-Reply-To field */
-    strfcpy (buffer, "In-Reply-To: ", sizeof (buffer));
-    mutt_make_string (buffer + 13, sizeof (buffer) - 13, InReplyTo, cur);
-    tmp = env->userhdrs;
-    while (tmp && tmp->next)
-      tmp = tmp->next;
-    if (tmp)
+    if (InReplyTo)
     {
-      tmp->next = mutt_new_list ();
-      tmp = tmp->next;
+      strfcpy (buffer, "In-Reply-To: ", sizeof (buffer));
+      mutt_make_string (buffer + 13, sizeof (buffer) - 13, InReplyTo, ctx, cur);
+      tmp = env->userhdrs;
+      while (tmp && tmp->next)
+	tmp = tmp->next;
+      if (tmp)
+      {
+	tmp->next = mutt_new_list ();
+	tmp = tmp->next;
+      }
+      else
+	tmp = env->userhdrs = mutt_new_list ();
+      tmp->data = safe_strdup (buffer);
     }
-    else
-      tmp = env->userhdrs = mutt_new_list ();
-    tmp->data = safe_strdup (buffer);
 
     env->references = mutt_copy_list (curenv->references);
 
@@ -623,7 +590,7 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
   else if (flags & SENDFORWARD)
   {
     /* set the default subject for the message. */
-    mutt_make_string (buffer, sizeof (buffer), ForwFmt, cur);
+    mutt_make_string (buffer, sizeof (buffer), ForwFmt, ctx, cur);
     env->subject = safe_strdup (buffer);
   }
 
@@ -670,7 +637,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
   }
   else if (flags & SENDFORWARD)
   {
-    if (option (OPTMIMEFWD))
+    if (query_quadoption (OPT_MIMEFWD, "Forward MIME encapsulated?"))
     {
       BODY *last = msg->content;
 
@@ -722,22 +689,26 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
   return (0);
 }
 
-static ADDRESS *mutt_set_followup_to (ENVELOPE *e)
+static void mutt_set_followup_to (ENVELOPE *e)
 {
   ADDRESS *t = NULL;
 
-  if (mutt_is_list_recipient (e->to) || mutt_is_list_recipient (e->cc))
+  /* only generate the Mail-Followup-To if the user has requested it, and
+     it hasn't already been set */
+  if (option (OPTFOLLOWUPTO) && !e->mail_followup_to)
   {
-    /* i am a list recipient, so set the Mail-Followup-To: field so that
-     * i don't end up getting multiple copies of responses to my mail
-     */
-    t = rfc822_append (&e->mail_followup_to, e->to);
-    rfc822_append (&t, e->cc);
-    /* the following is needed if $metoo is set, because the ->to and ->cc
-       may contain the user's private address(es) */
-    e->mail_followup_to = remove_user (e->mail_followup_to, 0);
+    if (mutt_is_list_recipient (e->to) || mutt_is_list_recipient (e->cc))
+    {
+      /* i am a list recipient, so set the Mail-Followup-To: field so that
+       * i don't end up getting multiple copies of responses to my mail
+       */
+      t = rfc822_append (&e->mail_followup_to, e->to);
+      rfc822_append (&t, e->cc);
+      /* the following is needed if $metoo is set, because the ->to and ->cc
+	 may contain the user's private address(es) */
+      e->mail_followup_to = remove_user (e->mail_followup_to, 0);
+    }
   }
-  return (e->mail_followup_to);
 }
 
 /* remove the multipart body if it exists */
@@ -755,6 +726,53 @@ static BODY *remove_multipart (BODY *b)
   return b;
 }
 
+/* look through the recipients of the message we are replying to, and if
+   we find an address that matches $alternates, we use that as the default
+   from field */
+static ADDRESS *set_reverse_name (ENVELOPE *env)
+{
+  ADDRESS *tmp;
+
+  for (tmp = env->to; tmp; tmp = tmp->next)
+  {
+    if (mutt_addr_is_user (tmp))
+      break;
+  }
+  if (!tmp)
+  {
+    for (tmp = env->cc; tmp; tmp = tmp->next)
+    {
+      if (mutt_addr_is_user (tmp))
+	break;
+    }
+  }
+  if (!tmp && mutt_addr_is_user (env->from))
+    tmp = env->from;
+  if (tmp)
+  {
+    tmp = rfc822_cpy_adr_real (tmp);
+    if (!tmp->personal)
+      tmp->personal = safe_strdup (Realname);
+  }
+  return (tmp);
+}
+
+static ADDRESS *mutt_default_from (void)
+{
+  ADDRESS *adr = rfc822_new_address ();
+
+  /* don't set realname here, it will be set later */
+
+  if (option (OPTUSEDOMAIN))
+  {
+    adr->mailbox = safe_malloc (strlen (Username) + strlen (Fqdn) + 2);
+    sprintf (adr->mailbox, "%s@%s", Username, Fqdn);
+  }
+  else
+    adr->mailbox = safe_strdup (Username);
+  return (adr);
+}
+
 void
 ci_send_message (int flags,		/* send mode */
 		 HEADER *msg,		/* template to use for new message */
@@ -766,10 +784,13 @@ ci_send_message (int flags,		/* send mode */
   char fcc[_POSIX_PATH_MAX] = ""; /* where to copy this message */
   FILE *tempfp = NULL;
   BODY *pbody;
-  int i;
-  int killfrom = 0;
+  int i, killfrom = 0;
+
+
 
    
+
+
   if (msg)
   {
     msg->env->to = mutt_expand_aliases (msg->env->to);
@@ -844,6 +865,19 @@ ci_send_message (int flags,		/* send mode */
     }
   }
 
+  /* this is handled here so that the user can match ~f in send-hook */
+  if (cur && option (OPTREVNAME))
+  {
+    /* we shouldn't have to worry about freeing `msg->env->from' before
+       setting it here since this code will only execute when doing some
+       sort of reply.  the pointer will only be set when using the -H command
+       line option */
+    msg->env->from = set_reverse_name (cur->env);
+  }
+
+  if (!msg->env->from && option (OPTUSEFROM))
+    msg->env->from = mutt_default_from ();
+
   if (flags & SENDBATCH) 
   {
     mutt_copy_stream (stdin, tempfp);
@@ -852,8 +886,6 @@ ci_send_message (int flags,		/* send mode */
       process_user_recips (msg->env);
       process_user_header (msg->env);
     }
-    if (option (OPTUSEFROM) && !msg->env->from)
-      msg->env->from = mutt_default_from ();
   }
   else if (! (flags & SENDPOSTPONED))
   {
@@ -879,8 +911,7 @@ ci_send_message (int flags,		/* send mode */
     if (!msg->env->from)
     {
       msg->env->from = mutt_default_from ();
-      if (!option (OPTUSEFROM))
-	killfrom = 1;
+      killfrom = 1;
     }
 
     /* change settings based upon recipients */
@@ -907,46 +938,23 @@ ci_send_message (int flags,		/* send mode */
     if (! (flags & (SENDMAILX | SENDKEY)) && strcmp (Editor, "builtin") != 0)
       append_signature (msg->env, tempfp);
   }
+  /* wait until now to set the real name portion of our return address so
+     that $realname can be set in a send-hook */
+  if (msg->env->from && !msg->env->from->personal)
+    msg->env->from->personal = safe_strdup (Realname);
+
+
 
     
+
+
     fclose (tempfp);
     tempfp = NULL;
     
-  if (cur && option (OPTREVNAME))
-  {
-    ADDRESS *tmp = NULL;
 
-    for (tmp = cur->env->to; tmp; tmp = tmp->next)
-    {
-      if (mutt_addr_is_user (tmp))
-	break;
-    }
 
-    if (!tmp)
-    {
-      for (tmp = cur->env->cc; tmp; tmp = tmp->next)
-      {
-	if (mutt_addr_is_user (tmp))
-	  break;
-      }
-    }
 
-    if (tmp)
-    {
-      ADDRESS *ptr;
 
-      rfc822_free_address (&msg->env->from);
-
-      /* rfc822_cpy_adr() copies the whole list, not just one element */
-      ptr = tmp->next;
-      tmp->next = NULL;
-      msg->env->from = rfc822_cpy_adr (tmp);
-      tmp->next = ptr;
-
-      if (!msg->env->from->personal)
-	msg->env->from->personal = safe_strdup (Realname);
-    }
-  }
 
   if (flags & SENDMAILX)
   {
@@ -993,7 +1001,9 @@ ci_send_message (int flags,		/* send mode */
     }
   }
 
-  if (!fcc[0])
+  /* specify a default fcc.  if we are in batchmode, only save a copy of
+     the message if the value of $copy is yes or ask-yes */
+  if (!fcc[0] && (!(flags & SENDBATCH) || (quadoption (OPT_COPY) & 0x1)))
   {
     /* set the default FCC */
     if (!msg->env->from)
@@ -1027,7 +1037,7 @@ main_loop:
       /* postpone the message until later. */
       if (msg->content->next)
 	msg->content = mutt_make_multipart (msg->content);
-      if (mutt_write_fcc (Postponed, msg, (cur && (flags & SENDREPLY)) ? cur->env->message_id : NULL, 1) < 0)
+      if (mutt_write_fcc (NONULL (Postponed), msg, (cur && (flags & SENDREPLY)) ? cur->env->message_id : NULL, 1) < 0)
       {
 	msg->content = remove_multipart (msg->content);
 	goto main_loop;
@@ -1080,18 +1090,20 @@ main_loop:
      * it by using an empty To: field.
      */
     msg->env->to = rfc822_new_address ();
-    msg->env->to->mailbox = safe_strdup ("undisclosed-recipients");
+    msg->env->to->mailbox = safe_strdup ("undisclosed-recipients:;");
     msg->env->to->group = 1;
     msg->env->to->next = rfc822_new_address ();
+ 
+    buffer[0] = 0;
+    rfc822_cat (buffer, sizeof (buffer), msg->env->to->mailbox, RFC822Specials);
+    msg->env->to->mailbox = safe_strdup (buffer);
   }
 
-  rfc822_free_address (&msg->env->mail_followup_to);
-  msg->env->mail_followup_to = mutt_set_followup_to (msg->env);
+  mutt_set_followup_to (msg->env);
 
   if (mutt_send_message (msg, fcc) == -1)
   {
     msg->content = remove_multipart (msg->content);
-    sleep (1); /* allow user to read the error message */
     goto main_loop;
   }
 
@@ -1113,12 +1125,20 @@ main_loop:
     }
   }
 
+
+
    
+
+
   return; /* all done */
 
 cleanup:
 
+
+
    
+
+
   if (tempfp)
     fclose (tempfp);
   mutt_free_header (&msg);
