@@ -17,6 +17,7 @@
  */ 
 
 #include "mutt.h"
+#include "mutt_curses.h"
 #include "rfc2047.h"
 #include "mx.h"
 #include "mime.h"
@@ -30,8 +31,9 @@
 #include <ctype.h>
 #include <sys/stat.h>
 #include <signal.h>
-
-
+#include <sys/wait.h>
+#include <fcntl.h>
+#include <sysexits.h>
 
 
 
@@ -404,6 +406,10 @@ int mutt_write_mime_body (BODY *a, FILE *f)
   }
 
 
+
+
+
+
   if ((fpin = fopen (a->filename, "r")) == NULL)
   {
     dprint(1,(debugfile, "write_mime_body: %s no longer exists!\n",a->filename));
@@ -621,7 +627,7 @@ static int lookup_mime_type (char *d, const char *s)
 
 static char *set_text_charset (CONTENT *info)
 {
-  if (strcasecmp (Charset, "us-ascii") == 0)
+  if (strcasecmp (NONULL (Charset), "us-ascii") == 0)
   {
     if (info->hibin != 0)
       return "unknown-8bit";
@@ -630,7 +636,7 @@ static char *set_text_charset (CONTENT *info)
     return "us-ascii";
 
   /* if no charset is given, provide a reasonable default */
-  return (Charset[0] ? Charset : "us-ascii");
+  return (Charset ? Charset : "us-ascii");
 }
 
 void mutt_message_to_7bit (BODY *a, FILE *fp)
@@ -816,6 +822,10 @@ void mutt_update_encoding (BODY *a)
   }
 
 
+
+
+
+
   safe_free ((void **) &info);
 }
 
@@ -864,6 +874,11 @@ BODY *mutt_make_attach (const char *path)
 
   mutt_set_encoding (att, info);
 
+
+
+
+
+
   safe_free ((void **) &info);
 
   return (att);
@@ -970,6 +985,8 @@ void mutt_write_address_list (ADDRESS *adr, FILE *fp, int linelen)
 /* arbitrary number of elements to grow the array by */
 #define REF_INC 16
 
+#define TrimRef 10
+
 /* need to write the list in reverse because they are stored in reverse order
  * when parsed to speed up threading
  */
@@ -1091,11 +1108,10 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
 #endif
 
   /* Add any user defined headers */
-  while (tmp)
+  for (; tmp; tmp = tmp->next)
   {
     fputs (tmp->data, fp);
     fputc ('\n', fp);
-    tmp = tmp->next;
   }
 
   return (ferror (fp) == 0 ? 0 : -1);
@@ -1157,34 +1173,342 @@ char *mutt_gen_msgid (void)
   return (safe_strdup (buf));
 }
 
-static int invoke_sendmail (const char *cmd, const char *tempfile)
+static RETSIGTYPE alarm_handler (int sig)
 {
-  char buf[LONG_STRING];
+  Signals |= S_ALARM;
+}
 
-  /* invoke sendmail in the background so the user doesn't have to wait for
-   * it to complete.  An example would be for someone who has dial-on-demand
-   * networking.
-   */
-  snprintf (buf, sizeof (buf), "%s < %s; /bin/rm -f %s", cmd, tempfile, tempfile);
-  return (_mutt_system (buf, M_DETACH_PROCESS));
+/* invoke sendmail in a subshell
+   path	(in)		path to program to execute
+   args	(in)		arguments to pass to program
+   msg (in)		temp file containing message to send
+   tempfile (out)	if sendmail is put in the background, this points
+   			to the temporary file containing the stdout of the
+			child process */
+static int
+send_msg (const char *path, char **args, const char *msg, char **tempfile)
+{
+  int fd, st, w = 0, err = 0;
+  pid_t pid;
+  struct sigaction act, oldint, oldquit, oldalrm;
+
+  memset (&act, 0, sizeof (struct sigaction));
+  sigemptyset (&(act.sa_mask));
+  act.sa_handler = SIG_IGN;
+  sigaction (SIGINT, &act, &oldint);
+  sigaction (SIGQUIT, &act, &oldquit);
+
+  if (SendmailWait)
+  {
+    char tmp[_POSIX_PATH_MAX];
+
+    mutt_mktemp (tmp);
+    *tempfile = safe_strdup (tmp);
+  }
+
+  if ((pid = fork ()) == 0)
+  {
+    /* reset signals for the child */
+    act.sa_handler = SIG_DFL;
+    /* we need SA_RESTART for the open() below */
+#ifdef SA_RESTART
+    act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
+#else
+    act.sa_flags = SA_NOCLDSTOP;
+#endif
+    sigaction (SIGCHLD, &act, NULL);
+    act.sa_flags = 0;
+    sigaction (SIGINT, &act, NULL);
+    sigaction (SIGQUIT, &act, NULL);
+
+    /* if it is possible that we will deliver in the background, then we have
+       to detach the child from this process group or it will die when the
+       parent process exists, causing the mail to not get delivered.  The
+       problem here is that any error messages will get lost... */
+    if (SendmailWait)
+      setsid ();
+    if ((pid = fork ()) == 0)
+    {
+      fd = open (msg, O_RDONLY, 0);
+      if (fd < 0)
+	_exit (127);
+      dup2 (fd, 0);
+      close (fd);
+      if (SendmailWait)
+      {
+	/* write stdout to a tempfile */
+	w = open (*tempfile, O_WRONLY | O_CREAT | O_EXCL, 0600);
+	if (w < 0)
+	  _exit (errno);
+	dup2 (w, 1);
+	close (w);
+      }
+      sigaction (SIGCHLD, &act, NULL);
+      execv (path, args);
+      unlink (msg);
+      _exit (127);
+    }
+    else if (pid == -1)
+    {
+      unlink (msg);
+      _exit (errno);
+    }
+
+    if (waitpid (pid, &st, 0) > 0)
+    {
+      st = WIFEXITED (st) ? WEXITSTATUS (st) : 127;
+      if (st == (EX_OK & 0xff) && SendmailWait)
+	unlink (*tempfile); /* no longer needed */
+    }
+    else
+    {
+      st = 127;
+      if (SendmailWait)
+	unlink (*tempfile);
+    }
+    unlink (msg);
+    _exit (st);
+  }
+  /* SendmailWait > 0: SIGALRM will interrupt wait() after alrm seconds
+     SendmailWait = 0: wait forever
+     SendmailWait < 0: don't wait */
+  if (SendmailWait > 0)
+  {
+    Signals &= ~S_ALARM;
+    act.sa_handler = alarm_handler;
+#ifdef SA_INTERRUPT
+    /* need to make sure the waitpid() is interrupted on SIGALRM */
+    act.sa_flags = SA_INTERRUPT;
+#else
+    act.sa_flags = 0;
+#endif
+    sigaction (SIGALRM, &act, &oldalrm);
+    alarm (SendmailWait);
+  }
+  if (SendmailWait >= 0)
+  {
+    w = waitpid (pid, &st, 0);
+    err = errno; /* save error since it might be clobbered by another
+		    system call before we check the value */
+    dprint (1, (debugfile, "send_msg(): waitpid returned %d (%s)\n", w,
+	    (w < 0 ? strerror (errno) : "OK")));
+  }
+  if (SendmailWait > 0)
+  {
+    alarm (0);
+    sigaction (SIGALRM, &oldalrm, NULL);
+  }
+  if (SendmailWait < 0 || ((Signals & S_ALARM) && w < 0 && err == EINTR))
+  {
+    /* add to list of children pids to reap */
+    mutt_add_child_pid (pid);
+    /* since there is no way for the user to be notified of error in this case,
+       remove the temp file now */
+    unlink (*tempfile);
+    FREE (tempfile);
+#ifdef DEBUG
+    if (Signals & S_ALARM)
+      dprint (1, (debugfile, "send_msg(): received SIGALRM\n"));
+    dprint (1, (debugfile, "send_msg(): putting sendmail in the background\n"));
+#endif
+    st = EX_OK & 0xff;
+  }
+  else if (w < 0)
+  {
+    /* if err==EINTR, alarm interrupt, child status unknown,
+       otherwise, there was an error invoking the child */
+    st = (err == EINTR) ? (EX_OK & 0xff) : 127;
+  }
+  else
+  {
+#ifdef DEBUG
+    if (WIFEXITED (st))
+    {
+      dprint (1, (debugfile, "send_msg(): child exited %d\n", WEXITSTATUS (st)));
+    }
+    else
+    {
+      dprint (1, (debugfile, "send_msg(): child did not exit\n"));
+    }
+#endif /* DEBUG */
+    st = WIFEXITED (st) ? WEXITSTATUS (st) : -1; /* return child status */
+  }
+  sigaction (SIGINT, &oldint, NULL);
+  sigaction (SIGQUIT, &oldquit, NULL);
+  /* restore errno so that the caller can build an error message */
+  errno = err;
+  return (st);
+}
+
+static char **
+add_args (char **args, size_t *argslen, size_t *argsmax, ADDRESS *addr)
+{
+  for (; addr; addr = addr->next)
+  {
+    /* weed out group mailboxes, since those are for display only */
+    if (addr->mailbox && !addr->group)
+    {
+      if (*argslen == *argsmax)
+	safe_realloc ((void **) &args, (*argsmax += 5) * sizeof (char *));
+      args[(*argslen)++] = addr->mailbox;
+    }
+  }
+  return (args);
+}
+
+static char **
+add_option (char **args, size_t *argslen, size_t *argsmax, char *s)
+{
+  if (*argslen == *argsmax)
+    safe_realloc ((void **) &args, (*argsmax += 5) * sizeof (char *));
+  args[(*argslen)++] = s;
+  return (args);
+}
+
+static int
+invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
+		 const char *msg, /* file containing message */
+		 int eightbit) /* message contains 8bit chars */
+{
+  char *ps = NULL, *path = NULL, *s = safe_strdup (Sendmail), *childout = NULL;
+  char **args = NULL;
+  size_t argslen = 0, argsmax = 0;
+  int i;
+
+  ps = s;
+  i = 0;
+  while ((ps = strtok (ps, " ")))
+  {
+    if (argslen == argsmax)
+      safe_realloc ((void **) &args, sizeof (char *) * (argsmax += 5));
+
+    if (i)
+      args[argslen++] = ps;
+    else
+    {
+      path = safe_strdup (ps);
+      ps = strrchr (ps, '/');
+      if (ps)
+	ps++;
+      else
+	ps = path;
+      args[argslen++] = ps;
+    }
+    ps = NULL;
+    i++;
+  }
+  if (eightbit && option (OPTUSE8BITMIME))
+    args = add_option (args, &argslen, &argsmax, "-B8BITMIME");
+  if (DsnNotify)
+  {
+    args = add_option (args, &argslen, &argsmax, "-N");
+    args = add_option (args, &argslen, &argsmax, DsnNotify);
+  }
+  if (DsnReturn)
+  {
+    args = add_option (args, &argslen, &argsmax, "-R");
+    args = add_option (args, &argslen, &argsmax, DsnReturn);
+  }
+  args = add_args (args, &argslen, &argsmax, to);
+  args = add_args (args, &argslen, &argsmax, cc);
+  args = add_args (args, &argslen, &argsmax, bcc);
+
+  if (argslen == argsmax)
+    safe_realloc ((void **) &args, sizeof (char *) * (++argsmax));
+  
+  args[argslen++] = NULL;
+
+  if (!option (OPTNOCURSES))
+    endwin ();
+  if ((i = send_msg (path, args, msg, &childout)) != (EX_OK & 0xff))
+  {
+    char *e = strerror (errno);
+
+    fprintf (stderr, "Error sending message, child exited %d (%s).\n", i, NONULL (e));
+    if (childout)
+      fprintf (stderr, "Saved output of child process to %s.\n", childout);
+    if (!option (OPTNOCURSES))
+    {
+      mutt_any_key_to_continue (NULL);
+      mutt_error ("Error sending message.");
+    }
+  }
+  FREE (&childout);
+  FREE (&path);
+  FREE (&s);
+  FREE (&args);
+
+  return (i);
+}
+
+/* appends string 'b' to string 'a', and returns the pointer to the new
+   string. */
+char *mutt_append_string (char *a, const char *b)
+{
+  size_t la = strlen (a);
+  safe_realloc ((void **) &a, la + strlen (b) + 1);
+  strcpy (a + la, b);
+  return (a);
+}
+
+/* returns 1 if char `c' needs to be quoted to protect from shell
+   interpretation when executing commands in a subshell */
+#define INVALID_CHAR(c) (!isalnum ((unsigned char)c) && !strchr ("@.+-_,:", c))
+
+/* returns 1 if string `s' contains characters which could cause problems
+   when used on a command line to execute a command */
+int mutt_needs_quote (const char *s)
+{
+  while (*s)
+  {
+    if (INVALID_CHAR (*s))
+      return 1;
+    s++;
+  }
+  return 0;
+}
+
+/* Quote a string to prevent shell escapes when this string is used on the
+   command line to send mail. */
+char *mutt_quote_string (const char *s)
+{
+  char *r, *pr;
+  size_t rlen;
+
+  rlen = strlen (s) + 3;
+  pr = r = malloc (rlen);
+  *pr++ = '"';
+  while (*s)
+  {
+    if (INVALID_CHAR (*s))
+    {
+      size_t o = pr - r;
+      safe_realloc ((void **) &r, ++rlen);
+      pr = r + o;
+      *pr++ = '\\';
+    }
+    *pr++ = *s++;
+  }
+  *pr++ = '"';
+  *pr = 0;
+  return (r);
 }
 
 int mutt_send_message (HEADER *msg, const char *fcc)
 {
-  char tempfile[_POSIX_PATH_MAX], buffer[STRING], cmd[LONG_STRING];
+  char tempfile[_POSIX_PATH_MAX], buffer[STRING];
   FILE *tempfp;
-  int len;
+  int i;
 
   /* Take care of 8-bit => 7-bit conversion. */
   rfc2047_encode_adrlist (msg->env->to);
   rfc2047_encode_adrlist (msg->env->cc);
-  rfc2047_encode_adrlist (msg->env->bcc);
   rfc2047_encode_adrlist (msg->env->from);
   rfc2047_encode_adrlist (msg->env->mail_followup_to);
   if (msg->env->subject)
   {
-    rfc2047_encode_string (buffer, sizeof (buffer) -1,
-			   (unsigned char *)msg->env->subject);
+    rfc2047_encode_string (buffer, sizeof (buffer) - 1,
+			   (unsigned char *) msg->env->subject);
     safe_free ((void **) &msg->env->subject);
     msg->env->subject = safe_strdup (buffer);
   }
@@ -1210,26 +1534,6 @@ int mutt_send_message (HEADER *msg, const char *fcc)
     return (-1);
   }
 
-  strfcpy (cmd, Sendmail, sizeof (cmd));
-  if (DsnNotify[0])
-  {
-    len = strlen (cmd);
-    snprintf (cmd + len, sizeof (cmd) - len, " -N %s", DsnNotify);
-  }
-  if (DsnReturn[0])
-  {
-    len = strlen (cmd);
-    snprintf (cmd + len, sizeof (cmd) - len, " -R %s", DsnReturn);
-  }
-
-  if (msg->content->encoding == ENC8BIT && option (OPTUSE8BITMIME))
-    strcat (cmd, " -B8BITMIME");
-
-  /* this should really check the return code, but at least under AIX,
-   * mutt_system() always returns -1
-   */
-  invoke_sendmail (cmd, tempfile);
-
   /* save a copy of the message, if necessary. */
   if (*fcc && strcmp ("/dev/null", fcc) != 0)
   {
@@ -1246,37 +1550,17 @@ int mutt_send_message (HEADER *msg, const char *fcc)
     msg->content = tmpbody;
   }
 
-  return 0;
-}
-
-static void write_addresses (char *s, size_t len, ADDRESS *a)
-{
-  size_t l;
-
-  while (a && len > 0)
-  {
-    l = strlen (a->mailbox);
-    if (l > len - 1)
-      l = len - 1;
-    memcpy (s, a->mailbox, l + 1); /* +1 to copy the  trailing \0 */
-    len -= l;
-    s += l;
-    a = a->next;
-
-    if (a && len > 0)
-    {
-      *s++ = ' ';
-      *s = 0;
-      len--;
-    }
-  }
+  i = invoke_sendmail (msg->env->to, msg->env->cc, msg->env->bcc,
+		       tempfile,
+		       (msg->content->encoding == ENC8BIT));
+  return (i ? -1 : 0);
 }
 
 void mutt_bounce_message (HEADER *h, ADDRESS *to)
 {
   int i;
   FILE *f;
-  char cmd[HUGE_STRING], date[SHORT_STRING], tempfile[_POSIX_PATH_MAX];
+  char date[SHORT_STRING], tempfile[_POSIX_PATH_MAX];
   MESSAGE *msg;
 
   if (!h)
@@ -1304,32 +1588,12 @@ void mutt_bounce_message (HEADER *h, ADDRESS *to)
       mutt_copy_bytes (msg->fp, f, h->content->length);
       fclose (f);
 
-      snprintf (cmd, sizeof (cmd), "%s ", SendmailBounce);
-      i = strlen (cmd);
-      write_addresses (cmd + i, sizeof (cmd) - i, to);
-      invoke_sendmail (cmd, tempfile);
+      invoke_sendmail (to, NULL, NULL, tempfile, h->content->encoding == ENC8BIT);
     }
-
     mx_close_message (&msg);
   }
 }
 
-ADDRESS *mutt_default_from (void)
-{
-  ADDRESS *adr = rfc822_new_address ();
-
-  adr->personal = safe_strdup (Realname);
-  
-  if (option (OPTUSEDOMAIN))
-  {
-    adr->mailbox = safe_malloc (strlen (Username) + strlen (Fqdn) + 2);
-    sprintf (adr->mailbox, "%s@%s", Username, Fqdn);
-  }
-  else
-    adr->mailbox = safe_strdup (Username);
-  return (adr);
-}
-
 /* given a list of addresses, return a list of unique addresses */
 ADDRESS *mutt_remove_duplicates (ADDRESS *addr)
 {
@@ -1418,7 +1682,11 @@ int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post)
     fprintf (msg->fp, "X-Mutt-References: %s\n", msgid);
   fprintf (msg->fp, "Status: RO\n");
 
-    
+
+
+
+
+
   if (tempfp)
   {
     char sasha[LONG_STRING];
